from __future__ import print_function, division, unicode_literals, absolute_import

import os
import tempfile
import unittest
import json


from monty.tempfile import ScratchDir
from fireworks import Workflow, Firework
from fireworks.core.rocket_launcher import rapidfire
from abiflows.core.testing import AbiflowsTest, has_mongodb
from abiflows.fireworks.tasks.tests import mock_objects
from abiflows.fireworks.tasks.utility_tasks import MongoEngineDBInsertionTask, FinalCleanUpTask, get_fw_task_manager
from abiflows.fireworks.utils.fw_utils import FWTaskManager
from abiflows.fireworks.tasks.tests.mock_objects import SaveDataWorkflow
from abiflows.database.mongoengine.utils import DatabaseData


MODULE_DIR = os.path.dirname(os.path.abspath(__file__))

test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "..", "..",
                        "test_files", "fw_task_managers")


class TestFinalCleanUpTask(AbiflowsTest):

    @classmethod
    def setUpClass(cls):
        cls.setup_fireworks()

    @classmethod
    def tearDownClass(cls):
        cls.teardown_fireworks(module_dir=MODULE_DIR)

    def tearDown(self):
        if self.lp:
            self.lp.reset(password=None,require_password=False)

    def test_class(self):
        """Test of basic methods"""

        task = FinalCleanUpTask()
        assert "WFK" in task.out_exts
        assert "1WF" in task.out_exts

        task = FinalCleanUpTask(out_exts=["DEN","WFK"])
        assert "WFK" in task.out_exts
        assert "DEN" in task.out_exts
        assert "1WF" not in task.out_exts

        task = FinalCleanUpTask(out_exts="DEN,WFK")
        assert "WFK" in task.out_exts
        assert "DEN" in task.out_exts
        assert "1WF" not in task.out_exts

        self.assertFwSerializable(task)

        with ScratchDir(".") as tmp_dir:

            wfkfile = tempfile.NamedTemporaryFile(suffix="_WFK", dir=tmp_dir)
            denfile = tempfile.NamedTemporaryFile(suffix="DEN", dir=tmp_dir)

            FinalCleanUpTask.delete_files(tmp_dir, exts=["WFK", "1WF"])

            assert not os.path.isfile(wfkfile.name)
            assert os.path.isfile(denfile.name)

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_run(self):
        create_fw = Firework([mock_objects.CreateOutputsTask(extensions=["WFK", "DEN"])], fw_id=1)
        delete_fw = Firework([FinalCleanUpTask(["WFK", "1WF"])], parents=create_fw, fw_id=2,
                             spec={"_add_launchpad_and_fw_id": True})

        wf = Workflow([create_fw, delete_fw])

        self.lp.add_wf(wf)

        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR, nlaunches=1)

        # check that the files have been created
        create_fw = self.lp.get_fw_by_id(1)
        create_ldir = create_fw.launches[0].launch_dir

        for d in ["tmpdata", "outdata", "indata"]:
            assert os.path.isfile(os.path.join(create_ldir, d, "tmp_WFK"))
            assert os.path.isfile(os.path.join(create_ldir, d, "tmp_DEN"))

        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR, nlaunches=1)

        wf = self.lp.get_wf_by_fw_id(1)

        assert wf.state == "COMPLETED"

        for d in ["tmpdata", "indata"]:
            assert not os.path.isfile(os.path.join(create_ldir, d, "tmp_WFK"))
            assert not os.path.isfile(os.path.join(create_ldir, d, "tmp_DEN"))

        assert not os.path.isfile(os.path.join(create_ldir, "outdata", "tmp_WFK"))
        assert os.path.isfile(os.path.join(create_ldir, "outdata", "tmp_DEN"))


class TestMongoEngineDBInsertionTask(AbiflowsTest):

    @classmethod
    def setUpClass(cls):
        cls.setup_fireworks()

    @classmethod
    def tearDownClass(cls):
        cls.teardown_fireworks(module_dir=MODULE_DIR)

    def test_class(self):
        db = DatabaseData("test_db", collection="test_collection", username="user", password="pass")
        task = MongoEngineDBInsertionTask(db)

        self.assertFwSerializable(task)

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_run(self):
        db = DatabaseData(self.lp.name, collection="test_MongoEngineDBInsertionTask", username=self.lp.username,
                          password=self.lp.password)
        task = MongoEngineDBInsertionTask(db)
        fw = Firework([task], fw_id=1, spec={"_add_launchpad_and_fw_id": True})
        wf = Workflow([fw], metadata={'workflow_class': SaveDataWorkflow.workflow_class,
                                     'workflow_module': SaveDataWorkflow.workflow_module})
        self.lp.add_wf(wf)

        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR, nlaunches=1)

        wf = self.lp.get_wf_by_fw_id(1)

        assert wf.state == "COMPLETED"

        # retrived the saved object
        # error if not imported locally
        from abiflows.fireworks.tasks.tests.mock_objects import DataDocument
        db.connect_mongoengine()
        with db.switch_collection(DataDocument) as DataDocument:
            data = DataDocument.objects()

            assert len(data) == 1

            assert data[0].test_field_string == "test_text"
            assert data[0].test_field_int == 5


class TestFunctions(AbiflowsTest):

    def test_get_fw_task_manager(self):
        manager_path = os.path.join(test_dir, "fw_manager_ok.yaml")

        fw_spec = {"ftm_file": manager_path, "fw_policy": {"max_restarts":44}}

        ftm = get_fw_task_manager(fw_spec)

        assert isinstance(ftm, FWTaskManager)
        assert ftm.fw_policy.max_restarts == 44
        assert ftm.has_task_manager()




from __future__ import print_function, division, unicode_literals, absolute_import

import os
import unittest

from abiflows.core.testing import AbiflowsTest, has_mongodb, TESTDB_NAME
from abiflows.database.mongoengine.utils import *
from mongoengine import Document, StringField
from mongoengine.context_managers import switch_db
from pymongo import MongoClient


class TestDatabaseData(AbiflowsTest):

    def setUp(self):
        self.setup_mongoengine()

    @classmethod
    def tearDownClass(cls):
        cls.teardown_mongoengine()

    def test_class(self):
        db_data = DatabaseData(TESTDB_NAME,host="localhost", port=27017, collection="test_collection", username="user",
                               password="password")

        self.assertMSONable(db_data)
        d = db_data.as_dict_no_credentials()
        assert "username" not in d
        assert "password" not in d

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_connection(self):
        db_data = DatabaseData(TESTDB_NAME, collection="test_collection")

        db_data.connect_mongoengine()

        class TestDocument(Document):
            test = StringField()

        with db_data.switch_collection(TestDocument) as TestDocument:
            TestDocument(test="abc").save()

        # check the collection with pymongo
        client = MongoClient()
        db = client[TESTDB_NAME]
        collection = db[db_data.collection]
        documents = collection.find()
        assert documents.count() == 1
        assert documents[0]['test'] == "abc"

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_connection_alias(self):

        test_db_name_2 = TESTDB_NAME+"_2"
        try:
            db_data2 = DatabaseData(test_db_name_2, collection="test_collection")
            db_data2.connect_mongoengine(alias="test_alias")

            db_data1 = DatabaseData(TESTDB_NAME, collection="test_collection")
            db_data1.connect_mongoengine()

            class TestDocument(Document):
                test = StringField()

            with switch_db(TestDocument, "test_alias") as TestDocument:
                with db_data2.switch_collection(TestDocument) as TestDocument:
                    TestDocument(test="abc").save()

            # check the collection with pymongo
            client = MongoClient()
            db = client[test_db_name_2]
            collection = db[db_data2.collection]
            documents = collection.find()
            assert documents.count() == 1
            assert documents[0]['test'] == "abc"
        finally:
            if self._connection:
                self._connection.drop_database(test_db_name_2)



from __future__ import print_function, division, unicode_literals, absolute_import

import pytest
import os
import glob
import unittest
import tempfile
import filecmp
import numpy.testing.utils as nptu
import abipy.data as abidata

from abipy.abio.factories import phonons_from_gsinput, PhononsFromGsFactory
from abipy.flowtk.tasks import TaskManager
from fireworks.core.rocket_launcher import rapidfire
from abiflows.fireworks.workflows.abinit_workflows import PhononFullFWWorkflow, PhononFWWorkflow
from abiflows.fireworks.utils.fw_utils import get_fw_by_task_index, load_abitask
from abiflows.core.testing import AbiflowsIntegrationTest, check_restart_task_type


ABINIT_VERSION = "8.6.1"

# pytestmark = [pytest.mark.skipif(not has_abinit(ABINIT_VERSION), reason="Abinit version {} is not in PATH".format(ABINIT_VERSION)),
#               pytest.mark.skipif(not has_fireworks(), reason="fireworks paackage is missing"),
#               pytest.mark.skipif(not has_mongodb(), reason="no connection to mongodb")]

# pytestmark = pytest.mark.usefixtures("cleandb")

class ItestPhonon(AbiflowsIntegrationTest):

    def itest_phonon_wf(self, lp, fworker, tmpdir, input_scf_phonon_si_low, use_autoparal, db_data):
        """
        Tests the complete running of PhononFullFWWorkflow and PhononFWWorkflow
        """

        # test at gamma. Pass a custom manager, to check proper serialization
        manager_path = os.path.join(abidata.dirpath, 'managers', "travis_manager.yml")
        ph_fac = PhononsFromGsFactory(qpoints=[[0,0,0]], ph_tol = {"tolvrs": 1.0e-7}, ddk_tol = {"tolwfr": 1.0e-16},
                                      dde_tol = {"tolvrs": 1.0e-7}, wfq_tol = {"tolwfr": 1.0e-16},
                                      manager=TaskManager.from_file(manager_path))

        # first run the phonon workflow with generation task
        wf_gen = PhononFWWorkflow(input_scf_phonon_si_low, ph_fac, autoparal=use_autoparal,
                                  initialization_info={"ngqpt": [1,1,1], "kppa": 100})

        wf_gen.add_anaddb_ph_bs_fw(input_scf_phonon_si_low.structure, ph_ngqpt=[1,1,1], ndivsm=2, nqsmall=2)
        wf_gen.add_mongoengine_db_insertion(db_data)
        wf_gen.add_final_cleanup(["WFK"])

        scf_id = wf_gen.scf_fw.fw_id
        ph_generation_fw_id = wf_gen.ph_generation_fw.fw_id
        old_new = wf_gen.add_to_db(lpad=lp)
        scf_id = old_new[scf_id]
        ph_generation_fw_id = old_new[ph_generation_fw_id]

        # run all the workflow
        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf_gen = lp.get_wf_by_fw_id(scf_id)

        assert wf_gen.state == "COMPLETED"

        ph_task = load_abitask(get_fw_by_task_index(wf_gen, "phonon_0", index=-1))

        # check the effect of the final cleanup
        assert len(glob.glob(os.path.join(ph_task.outdir.path, "*_WFK"))) == 0
        assert len(glob.glob(os.path.join(ph_task.outdir.path, "*_DEN1"))) > 0
        assert len(glob.glob(os.path.join(ph_task.tmpdir.path, "*"))) == 0
        assert len(glob.glob(os.path.join(ph_task.indir.path, "*"))) == 0

        # check the save in the DB
        from abiflows.database.mongoengine.abinit_results import PhononResult
        with db_data.switch_collection(PhononResult) as PhononResult:
            results = PhononResult.objects()
            assert len(results) == 1
            r = results[0]

            assert r.abinit_input.structure.to_mgobj() == input_scf_phonon_si_low.structure
            assert r.abinit_output.structure.to_mgobj() == input_scf_phonon_si_low.structure
            assert r.abinit_input.ecut == input_scf_phonon_si_low['ecut']
            assert r.abinit_input.kppa == 100
            nptu.assert_array_equal(r.abinit_input.gs_input.to_mgobj()['ngkpt'], input_scf_phonon_si_low['ngkpt'])
            nptu.assert_array_equal(r.abinit_input.ngqpt, [1,1,1])

            ana_task = load_abitask(get_fw_by_task_index(wf_gen, "anaddb", index=None))

            with tempfile.NamedTemporaryFile(mode="wb") as db_file:
                db_file.write(r.abinit_output.phonon_bs.read())
                db_file.seek(0)
                assert filecmp.cmp(ana_task.phbst_path, db_file.name)

            mrgddb_task = load_abitask(get_fw_by_task_index(wf_gen, "mrgddb", index=None))

            # read/write in binary for py3k compatibility with mongoengine
            with tempfile.NamedTemporaryFile(mode="wb") as db_file:
                db_file.write(r.abinit_output.ddb.read())
                db_file.seek(0)
                assert filecmp.cmp(mrgddb_task.merged_ddb_path, db_file.name)

        # then rerun a similar workflow, but completely generated at its creation
        wf_full = PhononFullFWWorkflow(input_scf_phonon_si_low, ph_fac, autoparal=use_autoparal)
        wf_full.add_anaddb_ph_bs_fw(input_scf_phonon_si_low.structure, ph_ngqpt=[1,1,1], ndivsm=2, nqsmall=2)
        wf_full.add_mongoengine_db_insertion(db_data)

        scf_id = wf_full.scf_fw.fw_id
        old_new = wf_full.add_to_db(lpad=lp)
        scf_id = old_new[scf_id]

        # run all the workflow
        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf_full = lp.get_wf_by_fw_id(scf_id)

        assert wf_full.state == "COMPLETED"

        # the full workflow doesn't contain the generation FW and the cleanup tasks, but should have the same
        # amount of perturbations.
        if use_autoparal:
            diff = 1
        else:
            diff = 2
        assert len(wf_full.id_fw) + diff == len(wf_gen.id_fw)

        if self.check_numerical_values:
            gen_scf_task = load_abitask(get_fw_by_task_index(wf_gen, "scf", index=-1))
            with gen_scf_task.open_gsr() as gen_gsr:
                gen_energy = gen_gsr.energy
                assert gen_energy == pytest.approx(-240.264972012, rel=0.01)

            gen_ana_task = load_abitask(get_fw_by_task_index(wf_gen, "anaddb", index=None))
            with gen_ana_task.open_phbst() as gen_phbst:
                gen_phfreq = gen_phbst.phbands.phfreqs[0, 3]
                assert gen_phfreq == pytest.approx(0.06029885, rel=0.1)

            full_scf_task = load_abitask(get_fw_by_task_index(wf_gen, "scf", index=-1))
            with full_scf_task.open_gsr() as full_gsr:
                full_energy = full_gsr.energy
                assert full_energy == pytest.approx(-240.264972012, rel=0.01)

            full_ana_task = load_abitask(get_fw_by_task_index(wf_gen, "anaddb", index=None))
            with full_ana_task.open_phbst() as full_phbst:
                full_phfreqs = full_phbst.phbands.phfreqs[0, 3]
                assert full_phfreqs == pytest.approx(0.06029885, rel=0.1)

            assert gen_energy == pytest.approx(full_energy, rel=1e-6)
            assert gen_phfreq == pytest.approx(full_phfreqs, rel=1e-6)

    def itest_not_converged(self, lp, fworker, tmpdir, input_scf_phonon_si_low):
        """
        Tests the missed convergence and restart for all the different kinds of tasks
        """

        # set a point not belonging to the grid so to trigger the calculation of WFQ and gamma for the DDE and DDK
        ph_inp = phonons_from_gsinput(input_scf_phonon_si_low, qpoints=[[0, 0, 0], [0.11111, 0.22222, 0.33333]],
                                      ph_tol = {"tolvrs": 1.0e-7}, ddk_tol = {"tolwfr": 1.0e-16},
                                      dde_tol = {"tolvrs": 1.0e-7}, wfq_tol = {"tolwfr": 1.0e-16})
        ph_inp.set_vars(nstep=3)
        wf_full = PhononFullFWWorkflow(input_scf_phonon_si_low, ph_inp, autoparal=False)

        scf_id = wf_full.scf_fw.fw_id
        old_new = wf_full.add_to_db(lpad=lp)
        scf_id = old_new[scf_id]

        # run the scf
        rapidfire(lp, fworker, m_dir=str(tmpdir), nlaunches=1)

        # pause all the remaining workflow and reignite the task types one by one to check the restart
        lp.pause_wf(scf_id)

        # DDK
        check_restart_task_type(lp, fworker, tmpdir, scf_id, "ddk_0")

        # reignite and run the other DDK to get to the DDE
        wf = lp.get_wf_by_fw_id(scf_id)
        lp.resume_fw(get_fw_by_task_index(wf, "ddk_1", index=1).fw_id)
        lp.resume_fw(get_fw_by_task_index(wf, "ddk_2", index=1).fw_id)
        rapidfire(lp, fworker, m_dir=str(tmpdir))

        # DDE
        check_restart_task_type(lp, fworker, tmpdir, scf_id, "dde_0")

        # NSCF
        check_restart_task_type(lp, fworker, tmpdir, scf_id, "nscf_wfq_0")

        # phonon
        check_restart_task_type(lp, fworker, tmpdir, scf_id, "phonon_0")

        # don't run the wf until the end to save time. Other tests covered that.
        wf = lp.get_wf_by_fw_id(scf_id)
        assert wf.state == "PAUSED"

    def itest_phonon_wfq_wf(self, lp, fworker, tmpdir, input_scf_phonon_si_low, db_data):
        """
        Tests the PhononFullFWWorkflow and PhononFWWorkflow
        """
        qpt = [[0.1111,0.2222,0.3333]]

        # test at gamma. Pass a custom manager, to check proper serialization
        manager_path = os.path.join(abidata.dirpath, 'managers', "travis_manager.yml")
        ph_fac = PhononsFromGsFactory(qpoints=qpt, ph_tol = {"tolvrs": 1.0e-7},
                                      wfq_tol = {"tolwfr": 1.0e-16}, with_ddk=False, with_dde=False,
                                      manager=TaskManager.from_file(manager_path))

        # first run the phonon workflow with generation task
        wf_gen = PhononFWWorkflow(input_scf_phonon_si_low, ph_fac, autoparal=False,
                                  initialization_info={"qpoints": qpt, "kppa": 100})

        wf_gen.add_mongoengine_db_insertion(db_data)

        scf_id = wf_gen.scf_fw.fw_id
        old_new = wf_gen.add_to_db(lpad=lp)
        scf_id = old_new[scf_id]

        # run all the workflow
        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf_gen = lp.get_wf_by_fw_id(scf_id)

        assert wf_gen.state == "COMPLETED"

        # then rerun a similar workflow, but completely generated at its creation
        wf_full = PhononFullFWWorkflow(input_scf_phonon_si_low, ph_fac, autoparal=False)

        scf_id = wf_full.scf_fw.fw_id
        old_new = wf_full.add_to_db(lpad=lp)
        scf_id = old_new[scf_id]

        # run all the workflow
        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf_full = lp.get_wf_by_fw_id(scf_id)

        assert wf_full.state == "COMPLETED"



#!/usr/bin/env python
"""Setup script for abiflows."""
from __future__ import print_function

import sys
import os
import shutil
import numpy as np

from glob import glob
from setuptools import find_packages, setup, Extension

if sys.version[0:3] < '2.7':
    sys.stderr.write("abiflows requires Python version 2.7 or above. Exiting.")
    sys.exit(1)

#-------------------------------------------------------------------------------
# Useful globals and utility functions
#-------------------------------------------------------------------------------

# A little utility we'll need below, since glob() does NOT allow you to do exclusion on multiple endings!
def file_doesnt_end_with(test, endings):
    """
    Returns true if test is a file and its name does NOT end with any
    of the strings listed in endings.
    """
    if not os.path.isfile(test):
        return False
    for e in endings:
        if test.endswith(e):
            return False
    return True

#---------------------------------------------------------------------------
# Basic project information
#---------------------------------------------------------------------------

# release.py contains version, authors, license, url, keywords, etc.
release_file = os.path.join('abiflows', 'core', 'release.py')

with open(release_file) as f:
    code = compile(f.read(), release_file, 'exec')
    exec(code)

#---------------------------------------------------------------------------
# Find package data
#---------------------------------------------------------------------------

def find_package_data():
    """Find abiflows package_data."""
    # This is not enough for these things to appear in an sdist.
    # We need to muck with the MANIFEST to get this to work
    package_data = {'abiflows.fireworks.tasks': ['n1000multiples_primes.json']
            }
    return package_data


def find_exclude_package_data():
    package_data = {
    #    'abiflows.data' : ["managers", 'benchmarks','runs/flow_*'],
    }
    return package_data


#---------------------------------------------------------------------------
# Find scripts
#---------------------------------------------------------------------------

def find_scripts():
    """Find abiflows scripts."""
    scripts = []
    # All python files in abiflows/scripts
    pyfiles = glob(os.path.join('abiflows', 'scripts', "*.py"))
    scripts.extend(pyfiles)
    return scripts


def get_long_desc():
    with open("README.md") as f:
        return f.read()


#-----------------------------------------------------------------------------
# Function definitions
#-----------------------------------------------------------------------------

def cleanup():
    """Clean up the junk left around by the build process."""

    if "develop" not in sys.argv:
        try:
            shutil.rmtree('abiflows.egg-info')
        except:
            try:
                os.unlink('abiflows.egg-info')
            except:
                pass

# List of external packages we rely on.
# Note setup install will download them from Pypi if they are not available.
#with open("requirements.txt", "rt") as fh:
#    install_requires = [s.strip() for s in fh]

install_requires = [
"six",
"numpy",
"pymongo",
"prettytable",
"mongoengine",
"paramiko",
"fireworks",
"abipy==0.6.0",
"custodian",
]

#---------------------------------------------------------------------------
# Find all the packages, package data, and data_files
#---------------------------------------------------------------------------

# Get the set of packages to be included.
my_packages = find_packages(exclude=())

my_scripts = find_scripts()

my_package_data = find_package_data()
my_excl_package_data = find_exclude_package_data()

# Create a dict with the basic information
# This dict is eventually passed to setup after additional keys are added.
setup_args = dict(
      name=name,
      version=version,
      description=description,
      long_description=long_description,
      author=author,
      author_email=author_email,
      url=url,
      license=license,
      platforms=platforms,
      keywords=keywords,
      classifiers=classifiers,
      install_requires=install_requires,
      packages=my_packages,
      package_data=my_package_data,
      exclude_package_data=my_excl_package_data,
      scripts=my_scripts,
      #download_url=download_url,
      )


if __name__ == "__main__":
    setup(**setup_args)
    cleanup()



# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

import os

from abiflows.core.testing import AbiflowsTest
from abiflows.fireworks.utils.time_utils import TimeReport, seconds_to_hms


class TestTimeReport(AbiflowsTest):
    def test_class(self):

        tr = TimeReport(120, 3, total_cpu_time=1200, contributed_cpu_time=3, total_run_time_per_tag={"tag": 10},
                 total_cpu_time_per_tag={"tag": 10}, contributed_cpu_time_per_tag={"tag": 3}, worker="worker name")

        self.assertMSONable(tr)

        str(tr)

        tr = TimeReport(120, 3)

        self.assertMSONable(tr)

        str(tr)


class TestFunctions(AbiflowsTest):

    def test_seconds_to_hms(self):
        time_string = seconds_to_hms(120)

        assert time_string == "0:02:00"



# coding: utf-8
"""
Utilities for fireworks
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import copy
import os
import traceback
import logging
import warnings
import json
import yaml
import io

from collections import namedtuple
from monty.json import MontyDecoder
from abipy.flowtk.tasks import TaskManager, ParalHints
from fireworks.core.firework import Firework, Workflow
from fireworks.core.launchpad import LaunchPad
from abiflows.fireworks.utils.time_utils import TimeReport

logger = logging.getLogger(__name__)

SHORT_SINGLE_CORE_SPEC = {'_queueadapter': {'ntasks': 1, 'time': '00:10:00'}, 'mpi_ncpus': 1}


def append_fw_to_wf(new_fw, wf):

    if new_fw.fw_id in wf.id_fw:
        raise ValueError('FW ids must be unique!')

    for leaf_id in wf.leaf_fw_ids:
        wf.links[leaf_id].append(new_fw.fw_id)

    wf.links[new_fw.fw_id] = []
    wf.id_fw[new_fw.fw_id] = new_fw

    # add depends on
    for pfw in new_fw.parents:
        if pfw.fw_id not in wf.links:
            raise ValueError(
                "FW_id: {} defines a dependent link to FW_id: {}, but the latter was not added to the workflow!".format(
                    new_fw.fw_id, pfw.fw_id))
        wf.links[pfw.fw_id].append(new_fw.fw_id)

    # sanity: make sure the set of nodes from the links_dict is equal to
    # the set of nodes from id_fw
    if set(wf.links.nodes) != set(map(int, wf.id_fw.keys())):
        raise ValueError("Specified links don't match given FW")

    wf.fw_states[new_fw.fw_id] = new_fw.state


def get_short_single_core_spec(fw_manager=None, master_mem_overhead=0, return_qtk=False, timelimit=None):
    if isinstance(fw_manager, FWTaskManager):
        ftm = fw_manager
    elif fw_manager:
        ftm = FWTaskManager.from_file(fw_manager)
    else:
        ftm = FWTaskManager.from_user_config()

    if ftm.has_task_manager():
        #TODO add mem_per_cpu?
        pconf = ParalHints({}, [{'tot_ncpus': 1, 'mpi_ncpus': 1, 'efficiency': 1}])
        try:
            tm = ftm.task_manager
            tm.select_qadapter(pconf)
            #TODO make a FW_task_manager parameter
            if timelimit is None:
                tm.qadapter.set_timelimit(timelimit=ftm.fw_policy.short_job_timelimit)
            else:
                tm.qadapter.set_timelimit(timelimit=timelimit)
            tm.qadapter.set_master_mem_overhead(master_mem_overhead)
            qadapter_spec = tm.qadapter.get_subs_dict()
            if return_qtk:
                return qadapter_spec, tm.qadapter
            else:
                return qadapter_spec
        except RuntimeError as e:
            traceback.print_exc()

    # No taskmanger or no queue available
    #FIXME return something else? exception?
    return {}


def set_short_single_core_to_spec(spec={}, master_mem_overhead=0, fw_manager=None):
        if spec is None:
            spec = {}
        else:
            spec = dict(spec)

        qadapter_spec = get_short_single_core_spec(master_mem_overhead=master_mem_overhead, fw_manager=fw_manager)
        spec['mpi_ncpus'] = 1
        spec['_queueadapter'] = qadapter_spec
        return spec


class FWTaskManager(object):
    """
    Object containing the configuration parameters and policies to run abipy.
    The policies needed for the abinit FW will always be available through default values. These can be overridden
    also setting the parameters in the spec.
    The standard abipy task manager is contained as an object on its own that can be used to run the autoparal or
    factories if needed.
    The rationale behind this choice, instead of subclassing, is to not force the user to fill the qadapter part
    of the task manager, which is needed only for the autoparal, but is required in the TaskManager initialization.
    Wherever the TaskManager is needed just pass the ftm.task_manager.
    The TaskManager part can be loaded from an external manager.yml file using the "abipy_manager" key in fw_policy.
    This is now the preferred choice. If this value is not defined, it will be loaded with TaskManager.from_user_config
    """

    YAML_FILE = "fw_manager.yaml"
    USER_CONFIG_DIR = TaskManager.USER_CONFIG_DIR # keep the same as the standard TaskManager

    fw_policy_defaults = dict(rerun_same_dir=False,
                              max_restarts=10,
                              autoparal=False,
                              abinit_cmd='abinit',
                              mrgddb_cmd='mrgddb',
                              anaddb_cmd='anaddb',
                              cut3d_cmd='cut3d',
                              mpirun_cmd='mpirun',
                              copy_deps=False,
                              walltime_command=None,
                              continue_unconverged_on_rerun=True,
                              allow_local_restart=False,
                              timelimit_buffer=120,
                              short_job_timelimit=600,
                              recover_previous_job=True,
                              abipy_manager=None)
    FWPolicy = namedtuple("FWPolicy", fw_policy_defaults.keys())

    def __init__(self, **kwargs):
        self._kwargs = copy.deepcopy(kwargs)

        fw_policy = kwargs.pop('fw_policy', {})
        unknown_keys = set(fw_policy.keys()) - set(self.fw_policy_defaults.keys())
        if unknown_keys:
            msg = "Unknown key(s) present in fw_policy: {}".format(", ".join(unknown_keys))
            logger.error(msg)
            raise RuntimeError(msg)
        fw_policy = dict(self.fw_policy_defaults, **fw_policy)

        # make a namedtuple for easier access to the attributes
        self.fw_policy = self.FWPolicy(**fw_policy)

        #TODO consider if raising an exception if it's requested when not available
        # create the task manager only if possibile
        if 'qadapters' in kwargs:
            self.task_manager = TaskManager.from_dict(kwargs)
            msg = "Loading the abipy TaskManager from inside the fw_manager.yaml file is deprecated. " \
                  "Use a separate file"
            logger.warning(msg)
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
        else:
            if self.fw_policy.abipy_manager:
                self.task_manager = TaskManager.from_file(self.fw_policy.abipy_manager)
            else:
                try:
                    self.task_manager = TaskManager.from_user_config()
                except:
                    logger.warning("Couldn't load the abipy task manager.")
                    self.task_manager = None

    @classmethod
    def from_user_config(cls, fw_policy=None):
        """
        Initialize the manager using the dict in the following order of preference:
        - the "fw_manager.yaml" file in the folder where the command is executed
        - a yaml file pointed by the "FW_TASK_MANAGER"
        - the "fw_manager.yaml" in the ~/.abinit/abipy folder
        - if no file available, fall back to default values
        """

        if fw_policy is None:
            fw_policy = {}

        # Try in the current directory then in user configuration directory.
        paths = [os.path.join(os.getcwd(), cls.YAML_FILE), os.getenv("FW_TASK_MANAGER"),
                 os.path.join(cls.USER_CONFIG_DIR, cls.YAML_FILE)]

        config = {}
        for path in paths:
            if path and os.path.exists(path):
                with io.open(path, "rt", encoding="utf-8") as fh:
                    config = yaml.load(fh)
                logger.info("Reading manager from {}.".format(path))
                break

        return cls(**config)

    @classmethod
    def from_file(cls, path):
        """Read the configuration parameters from the Yaml file filename."""
        with open(path, "rt") as fh:
            d = yaml.load(fh)
        return cls(**d)

    def has_task_manager(self):
        return self.task_manager is not None

    def update_fw_policy(self, d):
        self.fw_policy = self.fw_policy._replace(**d)


def get_time_report_for_wf(wf):

        total_run_time = 0
        total_cpu_time = 0
        contributed_cpu_time = 0
        total_run_time_per_tag = {}
        total_cpu_time_per_tag = {}
        contributed_cpu_time_per_tag = {}
        worker = None

        for fw in wf.fws:
            # skip not completed fws
            if fw.state != 'COMPLETED':
                continue

            launches = fw.archived_launches + fw.launches
            completed_launch = None

            #take the last completed launch
            for l in launches[-1::-1]:
                if l.state == 'COMPLETED':
                    completed_launch = l
                    break

            # assume all fw have runned on the same worker
            if worker is None:
                worker = completed_launch.fworker.name

            run_time = completed_launch.runtime_secs
            total_run_time += run_time

            if 'qtk_queueadapter' in fw.spec:
                ncpus = fw.spec['qtk_queueadapter'].num_cores
            elif 'mpi_ncpus' in fw.spec:
                ncpus = fw.spec['mpi_ncpus']
            elif 'ntasks' in fw.spec:
                ncpus = fw.spec['ntasks']
            else:
                ncpus = None

            if ncpus:
                cpu_time = run_time*ncpus
                total_cpu_time += cpu_time
                contributed_cpu_time += 1
            else:
                cpu_time = None

            wf_task_index = fw.spec.get('wf_task_index', 'Unclassified')

            task_tag = wf_task_index.rsplit('_', 1)[0]
            total_run_time_per_tag[task_tag] = total_run_time_per_tag.get(task_tag, 0) + run_time
            if cpu_time is not None:
                total_cpu_time_per_tag[task_tag] = total_cpu_time_per_tag.get(task_tag, 0) + cpu_time
                contributed_cpu_time_per_tag[task_tag] = contributed_cpu_time_per_tag.get(task_tag, 0) + 1

        tr = TimeReport(total_run_time, len(wf.fws), total_cpu_time=total_cpu_time, contributed_cpu_time=contributed_cpu_time,
                       total_run_time_per_tag=total_run_time_per_tag, total_cpu_time_per_tag=total_cpu_time_per_tag,
                       contributed_cpu_time_per_tag=contributed_cpu_time_per_tag, worker=worker)

        return tr


def links_dict_update(links_dict, links_update):
    for parent_id, child_ids in links_update.items():
        if isinstance(parent_id, Firework):
            parent_id = parent_id.fw_id
        if isinstance(child_ids, int):
            child_ids = [child_ids]
        elif isinstance(child_ids, Firework):
            child_ids = [child_ids.fw_id]
        if parent_id in links_dict:
            for child_id in child_ids:
                if child_id in links_dict[parent_id]:
                    raise ValueError('Child fireworks already defined for parent ...')
                links_dict[parent_id].append(child_id)
        else:
            links_dict[parent_id] = child_ids


def get_last_completed_launch(fw):
    """
    Given a Firework object returns the last completed launch
    """
    return next((l for l in reversed(fw.archived_launches + fw.launches) if
                 l.state == 'COMPLETED'), None)


def load_abitask(fw):
    """
    Given a Firework object returns the abinit related task contained. Sets the list of directories set from the
    last completed launch. If no abinit related firetasks are found or the task has no completed launch returns None.
    """

    from abiflows.fireworks.tasks.abinit_tasks import AbiFireTask, AnaDdbAbinitTask, MergeDdbAbinitTask

    for t in fw.tasks:
        if isinstance(t, (AbiFireTask, AnaDdbAbinitTask, MergeDdbAbinitTask)):
            launch = get_last_completed_launch(fw)
            if launch:
                t.set_workdir(workdir=launch.launch_dir)
                return t

    return None

def get_fw_by_task_index(wf, task_tag, index=1):
    """
    Given a workflow object (with connection to the db) returns the wf corresponding to the task_type.

    Args:
        wf: a fireworks Workflow object.
        task_tag: the task tag associated with the task as defined in abinit_workflows. Should not include the index.
        index: the numerical or text index of the task. If negative the the last fw corresponding to task_tag will
            be selected. If None, no index will be considered and the first match will be returned.

    Returns:
        a fireworks Firework object. None if no match is found.
    """

    task_index = None
    if index is not None and index >=0:
        task_index = "{}_{}".format(task_tag, index)

    selected_fw = None
    max_ind = -1
    for fw in wf.fws:
        fw_task_index = fw.spec.get('wf_task_index', '')
        if task_index:
            if fw_task_index == task_index:
                return fw
        else:
            if task_tag in fw_task_index:
                if index is None:
                    return fw
                # the last part of the task_index can be text (i.e. "autoparal") so the conversion to int may fail
                # if no other indices has been found select that one
                try:
                    fw_ind = int(fw_task_index.split('_')[-1])
                    if  fw_ind > max_ind:
                        selected_fw = fw
                        max_ind = fw_ind
                except:
                    if selected_fw is None:
                        selected_fw = fw

    return selected_fw


def get_lp_and_fw_id_from_task(task, fw_spec):
    """
    Given an instance of a running task and its spec, tries to load the LaunchPad and the current fw_id.
    It will first check for "_add_launchpad_and_fw_id", then try to load from FW.json/FW.yaml file.

    Should be used inside tasks that require to access to the LaunchPad and to the whole workflow.
    Args:
        task: An instance of a running task
        fw_spec: The spec of the task

    Returns:
        an instance of LaunchPah and the fw_id of the current task
    """
    if '_add_launchpad_and_fw_id' in fw_spec:
        lp = task.launchpad
        fw_id = task.fw_id

        # lp may be None in offline mode
        if lp is None:
            raise RuntimeError("The LaunchPad in spec is None.")
    else:
        try:
            with open('FW.json', "rt") as fh:
                fw_dict = json.load(fh, cls=MontyDecoder)
        except IOError:
            try:
                with open('FW.yaml', "rt") as fh:
                    fw_dict = yaml.load(fh)
            except IOError:
                raise RuntimeError("Launchpad/fw_id not present in spec and No FW.json nor FW.yaml file present: "
                                   "impossible to determine fw_id")

        logger.warning("LaunchPad not available from spec. Generated with auto_load.")
        lp = LaunchPad.auto_load()
        fw_id = fw_dict['fw_id']

        # since it is not given that the LaunchPad is the correct one, try to verify if the workflow
        # and the fw_id are being accessed correctly
        try:
            fw = lp.get_fw_by_id(fw_id)
        except ValueError as e:
            traceback.print_exc()
            raise RuntimeError("The firework with id {} is not present in the LaunchPad {}. The LaunchPad is "
                               "probably incorrect.". format(fw_id, lp))

        if fw.state != "RUNNING":
            raise RuntimeError("The firework with id {} from LaunchPad {} is {}. There might be an error in the "
                               "selection of the LaunchPad". format(fw_id, lp, fw.state))

        if len(fw.tasks) != len(fw_dict['spec']['_tasks']):
            raise RuntimeError("The firework with id {} from LaunchPad {} is has different number of tasks "
                               "from the current.".format(fw_id, lp))

        for db_t, dict_t in zip(fw.tasks, fw_dict['spec']['_tasks']):
            if db_t.fw_name != dict_t['_fw_name']:
                raise RuntimeError("The firework with id {} from LaunchPad {} has task that don't  match: "
                                   "{} and {}.".format(fw_id, lp, db_t.fw_name, dict_t['fw_name']))

    return lp, fw_id



from __future__ import print_function, division, unicode_literals, absolute_import

import os

from pymatgen.io.abinit.events import EventReport, ScfConvergenceWarning, RelaxConvergenceWarning, AbinitError

from fireworks import Firework, FireTaskBase, FWAction, explicit_serialize
from abiflows.fireworks.workflows.abinit_workflows import AbstractFWWorkflow
from mongoengine import Document, StringField, IntField

##########################
# Test reports
##########################

def report_ok():
    return EventReport('.')

def report_ScfConvergenceWarning():
    er = EventReport('.', events=[ScfConvergenceWarning(message='Fake warning', src_file=__file__, src_line=0)])
    er.set_run_completed(True,'', ' ')
    return er

def report_RelaxConvergenceWarning():
    return EventReport('.', events=[RelaxConvergenceWarning(message='Fake warning', src_file=__file__, src_line=0)])

def report_AbinitError():
    return EventReport('.', events=[AbinitError(message='Fake warning', src_file=__file__, src_line=0)])

##########################
# Fake Tasks
##########################

@explicit_serialize
class FakeTask(FireTaskBase):
    def run_task(self, fw_spec):
        return FWAction()

fake_fw = Firework([FakeTask()])


@explicit_serialize
class CreateOutputsTask(FireTaskBase):
    """
    Creates temporary files in with the specified extensions in the "indata", "outdata" and "tmpdata" folders.
    """

    prefix = "tmp_"

    def run_task(self, fw_spec):
        dirs = ["indata", "outdata", "tmpdata"]
        for d in dirs:
            os.mkdir(d)
            for e in self.get('extensions', []):
                with open(os.path.join(d, "{}{}".format(self.prefix, e)), "wt") as f:
                    f.write(" ")
        return FWAction()


##########################
# Fake Workflows
##########################

class DataDocument(Document):

    test_field_string = StringField()
    test_field_int = IntField()

class SaveDataWorkflow(AbstractFWWorkflow):

    workflow_class = 'SaveDataWorkflow'
    workflow_module = 'abiflows.fireworks.tasks.tests.mock_objects'

    @classmethod
    def get_mongoengine_results(cls, wf):
        return DataDocument(test_field_string="test_text", test_field_int=5)




# coding: utf-8
"""
Common test support for abiflows test scripts.

This single module should provide all the common functionality for abiflows tests
in a single location, so that test scripts can just import it and work right away.
This module heavily depends on the abipy.testing module/
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import numpy
import subprocess
import json
import tempfile
import shutil
import glob
import unittest
import numpy.testing.utils as nptu

from mongoengine import connect, Document
from mongoengine.connection import get_db, get_connection
from monty.os.path import which
from monty.string import is_string
from abipy.core.testing import AbipyTest
from fireworks.core.launchpad import LaunchPad
from fireworks.core.fworker import FWorker
from fireworks.core.rocket_launcher import rapidfire
from abiflows.fireworks.utils.fw_utils import get_fw_by_task_index

import logging
logger = logging.getLogger(__file__)

root = os.path.dirname(__file__)

__all__ = [
    "AbiflowsTest"
]


TESTDB_NAME = "abiflows_unittest"


def has_mongodb(host='localhost', port=27017, name='mongodb_test', username=None, password=None):
   try:
       from pymongo import MongoClient
       connection = MongoClient(host, port, j=True)
       db = connection[name]
       if username:
           db.authenticate(username, password)

       return True
   except:
       return False


def has_fireworks():
   """True if fireworks is installed."""
   try:
       import fireworks
       return True
   except ImportError:
       return False


class AbiflowsTest(AbipyTest):
    """Extends AbipyTest with methods specific for the testing of workflows"""

    def assertFwSerializable(self, obj):
       assert '_fw_name' in obj.to_dict()
       self.assertDictEqual(obj.to_dict(), obj.__class__.from_dict(obj.to_dict()).to_dict())

    @classmethod
    def setup_fireworks(cls):
        """
        Sets up the fworker and launchpad if a connection to a local mongodb is available.
        cls.lp is set to None if not available
        """

        cls.fworker = FWorker()
        try:
            cls.lp = LaunchPad(name=TESTDB_NAME, strm_lvl='ERROR')
            cls.lp.reset(password=None, require_password=False)
        except:
            cls.lp = None

    @classmethod
    def teardown_fireworks(cls, module_dir=None):
        """
        Removes the fireworks test database if cls.lp is present and deletes all the launcher directories
        """
        if cls.lp:
            cls.lp.connection.drop_database(TESTDB_NAME)

        if module_dir:
            for ldir in glob.glob(os.path.join(module_dir,"launcher_*")):
                shutil.rmtree(ldir)

    @classmethod
    def setup_mongoengine(cls):
        try:
            cls._connection = connect(db=TESTDB_NAME)
            cls._connection.drop_database(TESTDB_NAME)
            cls.db = get_db()
        except:
            cls.db = None
            cls._connection = None

    @classmethod
    def teardown_mongoengine(cls):
        if cls._connection:
            cls._connection.drop_database(TESTDB_NAME)


    def get_document_class_from_mixin(self, mixin_cls):
        """
        Utility function to generate a mongoengine Document class from the mixin.
        Needed to save the object in the db with mongoengine
        """

        class TestDocument(mixin_cls, Document):
            meta = {'collection': "test_{}".format(mixin_cls.__name__)}

        return TestDocument


class AbiflowsIntegrationTest(object):
    """
    Provides utility methods and variables for integration tests, that can't subclass unittest.TestCase
    """

    # variable to enable/disable the checks on the numerical quantities as output of the workflow
    check_numerical_values = True

    @staticmethod
    def assertArrayAlmostEqual(actual, desired, decimal=7, err_msg='',
                               verbose=True):
        """
        Tests if two arrays are almost equal to a tolerance. The CamelCase
        naming is so that it is consistent with standard unittest methods.
        """
        return nptu.assert_almost_equal(actual, desired, decimal, err_msg, verbose)


def check_restart_task_type(lp, fworker, tmpdir, fw_id, task_tag):

    # resume the task for tag
    wf = lp.get_wf_by_fw_id(fw_id)
    fw = get_fw_by_task_index(wf, task_tag, index=1)
    assert fw is not None
    assert fw.state == "PAUSED"
    lp.resume_fw(fw.fw_id)

    # run the FW
    rapidfire(lp, fworker, m_dir=str(tmpdir), nlaunches=1)

    # the job should have a detour for the restart
    wf = lp.get_wf_by_fw_id(fw_id)
    fw = get_fw_by_task_index(wf, task_tag, index=2)
    assert fw is not None
    assert fw.state == "READY"

    # run all the following and check that the last is correctly completed (if convergence is not achieved
    # the final state should be FIZZLED)
    rapidfire(lp, fworker, m_dir=str(tmpdir))

    wf = lp.get_wf_by_fw_id(fw_id)
    fw = get_fw_by_task_index(wf, task_tag, index=-1)

    assert fw.state == "COMPLETED"



from __future__ import print_function, division, unicode_literals, absolute_import

import abc
import copy
import inspect
import json
import os
import six

from fireworks.core.firework import FireTaskBase
from fireworks.core.firework import FWAction
from fireworks.core.firework import Workflow
from fireworks.core.firework import Firework
from fireworks.core.launchpad import LaunchPad
from fireworks.utilities.fw_utilities import explicit_serialize
from fireworks.utilities.fw_serializers import serialize_fw

from monty.json import MontyDecoder, MontyEncoder
from monty.json import MSONable
from monty.serialization import loadfn
from monty.subprocess import Command

from abiflows.core.mastermind_abc import ControlProcedure, ControlledItemType
from abiflows.core.mastermind_abc import ControllerNote
from abiflows.core.mastermind_abc import Cleaner
from abiflows.fireworks.utils.fw_utils import set_short_single_core_to_spec, get_short_single_core_spec

RESTART_FROM_SCRATCH = ControllerNote.RESTART_FROM_SCRATCH
RESET_RESTART = ControllerNote.RESET_RESTART
SIMPLE_RESTART = ControllerNote.SIMPLE_RESTART

class SRCTaskMixin(object):

    src_type = ''

    @serialize_fw
    def to_dict(self):
        d = {}
        for arg in inspect.getargspec(self.__init__).args:
            if arg != "self":
                val = self.__getattribute__(arg)
                if hasattr(val, "as_dict"):
                    val = val.as_dict()
                elif isinstance(val, (tuple, list)):
                    val = [v.as_dict() if hasattr(v, "as_dict") else v for v in val]
                d[arg] = val

        return d

    @classmethod
    def from_dict(cls, d):
        dec = MontyDecoder()
        kwargs = {k: dec.process_decoded(v) for k, v in d.items()
                  if k in inspect.getargspec(cls.__init__).args}
        return cls(**kwargs)

    def setup_directories(self, fw_spec, create_dirs=False):
        if self.src_type == 'setup':
            self.src_root_dir = fw_spec.get('_launch_dir', os.getcwd())
        elif 'src_directories' in fw_spec:
            self.src_root_dir = fw_spec['src_directories']['src_root_dir']
        # elif self.src_type in ['run', 'control']:
        #     self.src_root_dir = os.path.split(os.path.abspath(fw_spec['_launch_dir']))[0]
        else:
            raise ValueError('Cannot setup directories for "src_type" = "{}"'.format(self.src_type))
        self.setup_dir = os.path.join(self.src_root_dir, 'setup')
        self.run_dir = os.path.join(self.src_root_dir, 'run')
        self.control_dir = os.path.join(self.src_root_dir, 'control')
        # if 'src_directories' in fw_spec:
        #     if (self.src_root_dir != fw_spec['src_directories']['src_root_dir'] or
        #         self.setup_dir != fw_spec['src_directories']['setup_dir'] or
        #         self.run_dir != fw_spec['src_directories']['run_dir'] or
        #         self.control_dir != fw_spec['src_directories']['control_dir']):
        #         raise ValueError('src_directories in fw_spec do not match actual SRC directories ...')
        if create_dirs:
            os.makedirs(self.setup_dir)
            os.makedirs(self.run_dir)
            os.makedirs(self.control_dir)

    @property
    def src_directories(self):
        return {'src_root_dir': self.src_root_dir,
                'setup_dir': self.setup_dir,
                'run_dir': self.run_dir,
                'control_dir': self.control_dir
                }


@explicit_serialize
class SetupTask(SRCTaskMixin, FireTaskBase):

    src_type = 'setup'
    task_type = 'unknown'

    RUN_PARAMETERS = ['_queueadapter', 'mpi_ncpus', 'qtk_queueadapter']

    def __init__(self, deps=None, restart_info=None, task_type=None):

        # TODO: if at some point, this is not enough (as for the ddk, or for the structure, or for anything else,
        #       we could thing of an object ?
        # deps are transformed to be a list or a dict of lists
        if isinstance(deps, dict):
            deps = dict(deps)
            for k, v in deps.items():
                if not isinstance(v, (list, tuple)):
                    deps[k] = [v]
        elif deps and not isinstance(deps, (list, tuple)):
            deps = [deps]
        self.deps = deps

        self.restart_info = restart_info

        if task_type is not None:
            self.task_type = task_type

    def set_restart_info(self, restart_info=None):
        self.restart_info = restart_info

    def run_task(self, fw_spec):
        # Set up and create the directory tree of the Setup/Run/Control trio,
        self.setup_directories(fw_spec=fw_spec, create_dirs=True)
        #  Forward directory information to run and control fireworks #HACK in _setup_run_and_control_dirs
        self._setup_run_and_control_dirs_and_fworker(fw_spec=fw_spec)
        # Move to the setup directory
        os.chdir(self.setup_dir)
        # Make the file transfers from another worker if needed
        self.file_transfers(fw_spec=fw_spec)
        # Get back information from the previous runs
        self.fetch_previous_info(fw_spec=fw_spec)
        # Setup the parameters for the run (number of cpus, time, memory, openmp, ...)
        params = list(self.RUN_PARAMETERS)
        if 'src_modified_objects' in fw_spec:
            for target, modified_object in fw_spec['src_modified_objects'].items():
                params.remove(target)
        run_parameters = self._setup_run_parameters(fw_spec=fw_spec, parameters=params)
        # Prepare run (make links to output files from previous tasks, write input files, create the directory
        # tree of the program, ...)
        self.prepare_run(fw_spec=fw_spec)

        # Update the spec of the Run Firework with the directory tree, the run_parameters obtained from
        #  setup_run_parameters and the modified_objects transferred directly from the Control Firework
        update_spec = {'src_directories': self.src_directories}
        update_spec.update(run_parameters)
        if 'src_modified_objects' in fw_spec:
            update_spec.update(fw_spec['src_modified_objects'])
        if 'previous_fws' in fw_spec:
            update_spec['previous_fws'] = fw_spec['previous_fws']
        return FWAction(update_spec=update_spec)

    def _setup_run_parameters(self, fw_spec, parameters):
        qadapter_spec, qtk_queueadapter = get_short_single_core_spec(return_qtk=True)
        params = {'_queueadapter': qadapter_spec, 'mpi_ncpus': 1, 'qtk_queueadapter': qtk_queueadapter}
        setup_params = self.setup_run_parameters(fw_spec=fw_spec)
        params.update(setup_params)
        if 'initial_parameters' in fw_spec and fw_spec['SRC_task_index'].index == 1:
            qtk_queueadapter = params['qtk_queueadapter']
            initial_parameters = fw_spec['initial_parameters']
            if 'run_timelimit' in initial_parameters:
                qtk_queueadapter.set_timelimit(timelimit=initial_parameters['run_timelimit'])
            if 'run_mem_per_proc' in initial_parameters:
                qtk_queueadapter.set_mem_per_proc(mem_mb=initial_parameters['run_mem_per_proc'])
            if 'run_mpi_ncpus' in initial_parameters:
                qtk_queueadapter.set_mpi_procs(mpi_procs=initial_parameters['run_mpi_ncpus'])
            qadapter_spec = qtk_queueadapter.get_subs_dict()
            params.update({'qtk_queueadapter': qtk_queueadapter, '_queueadapter': qadapter_spec})
        return {param: params[param] for param in parameters}

    def setup_run_parameters(self, fw_spec):
        return {}

    def file_transfers(self, fw_spec):
        pass

    def fetch_previous_info(self, fw_spec):
        pass

    def prepare_run(self, fw_spec):
        pass

    def _setup_run_and_control_dirs_and_fworker(self, fw_spec):
        """
        This method is used to update the spec of the run and control fireworks with the src_directories as well as
        set the _launch_dir of the run and control fireworks to be the run_dir and control_dir respectively.
        WARNING: This is a bit hackish! Do not change this unless you know exactly what you are doing!
        :param fw_spec: Firework's spec
        """
        # Get the launchpad
        if '_add_launchpad_and_fw_id' in fw_spec:
            lp = self.launchpad
            setup_fw_id = self.fw_id
        else:
            try:
                fw_dict = loadfn('FW.json')
            except IOError:
                try:
                    fw_dict = loadfn('FW.yaml')
                except IOError:
                    raise RuntimeError("Launchpad/fw_id not present in spec and No FW.json nor FW.yaml file present: "
                                       "impossible to determine fw_id")
            lp = LaunchPad.auto_load()
            setup_fw_id = fw_dict['fw_id']
        if '_add_fworker' in fw_spec:
            fworker = self.fworker
        else:
            raise ValueError('Should have access to the fworker in SetupTask ...')
        this_lzy_wf = lp.get_wf_by_fw_id_lzyfw(setup_fw_id)
        # Check that SetupTask and RunTask have only one child firework
        child_fw_ids = this_lzy_wf.links[setup_fw_id]
        if len(child_fw_ids) != 1:
            raise ValueError('SetupTask\'s Firework should have exactly one child firework')
        run_fw_id = child_fw_ids[0]
        child_run_fw_ids = this_lzy_wf.links[run_fw_id]
        if len(child_run_fw_ids) != 1:
            raise ValueError('RunTask\'s Firework should have exactly one child firework')
        control_fw_id = child_run_fw_ids[0]
        spec_update = {'_launch_dir': self.run_dir,
                       'src_directories': self.src_directories,
                       '_fworker': fworker.name}
        lp.update_spec(fw_ids=[run_fw_id],
                       spec_document=spec_update)
        spec_update['_launch_dir'] = self.control_dir
        lp.update_spec(fw_ids=[control_fw_id],
                       spec_document=spec_update)

    def additional_task_info(self):
        return {}


class RunTask(SRCTaskMixin, FireTaskBase):

    src_type = 'run'
    task_type = 'unknown'

    def __init__(self, control_procedure, task_type=None):
        self.set_control_procedure(control_procedure=control_procedure)
        if task_type is not None:
            self.task_type = task_type

    def set_control_procedure(self, control_procedure):
        self.control_procedure = control_procedure
        #TODO: check something here with the monitors ?

    def run_task(self, fw_spec):
        self.setup_directories(fw_spec=fw_spec, create_dirs=False)
        launch_dir = os.getcwd()
        # Move to the run directory
        os.chdir(self.run_dir)
        f = open(os.path.join(self.run_dir, 'fw_info.txt'), 'a')
        f.write('FW launch_directory :\n{}'.format(launch_dir))
        f.close()
        # The Run and Control tasks have to run on the same worker

        #TODO: do something here with the monitoring controllers ...
        #      should stop the RunTask but the correction should be applied in control !
        self.config(fw_spec=fw_spec)
        self.run(fw_spec=fw_spec)
        update_spec = self.postrun(fw_spec=fw_spec)

        if update_spec is None:
            update_spec = {}

        if 'previous_fws' in fw_spec:
            update_spec['previous_fws'] = fw_spec['previous_fws']

        #TODO: the directory is passed thanks to _pass_job_info. Should we pass anything else ?
        return FWAction(stored_data=None, exit=False, update_spec=None, mod_spec=None,
                        additions=None, detours=None,
                        defuse_children=False)

    def config(self, fw_spec):
        pass

    @abc.abstractmethod
    def run(self, fw_spec):
        pass

    def postrun(self, fw_spec):
        pass

    def additional_task_info(self):
        return {}


@explicit_serialize
class ScriptRunTask(RunTask):

    task_type = 'script'

    def __init__(self, script_str, control_procedure):
        RunTask.__init__(self, control_procedure=control_procedure)
        self.script_str = script_str

    def run(self, fw_spec):
        f = open('script_run.log', 'w')
        cmds_strs = self.script_str.split(';')
        for cmd_str in cmds_strs:
            cmd = Command(cmd_str)
            cmd = cmd.run()
            if cmd.retcode != 0:
                raise ValueError('Command "{}" returned exit code {:d}'.format(cmd_str, cmd.retcode))
            if cmd.output is not None:
                print(cmd.output)
            f.write('{}\n'.format(str(cmd)))
        f.close()

    @serialize_fw
    def to_dict(self):
        return {'script_str': self.script_str,
                'control_procedure': self.control_procedure.as_dict()}

    @classmethod
    def from_dict(cls, d):
        control_procedure = ControlProcedure.from_dict(d['control_procedure'])
        return cls(script_str=d['script_str'], control_procedure=control_procedure)


@explicit_serialize
class ControlTask(SRCTaskMixin, FireTaskBase):
    src_type = 'control'

    def __init__(self, control_procedure, manager=None, max_restarts=10, src_cleaning=None):
        self.control_procedure = control_procedure
        self.manager = manager
        self.max_restarts = max_restarts
        self.src_cleaning = src_cleaning

    def run_task(self, fw_spec):
        self.setup_directories(fw_spec=fw_spec, create_dirs=False)
        launch_dir = os.getcwd()
        # Move to the control directory
        os.chdir(self.control_dir)
        f = open(os.path.join(self.control_dir, 'fw_info.txt'), 'a')
        f.write('FW launch_directory :\n{}'.format(launch_dir))
        f.close()
        # Get the task index
        task_index = SRCTaskIndex.from_any(fw_spec['SRC_task_index'])
        # Get the setup and run fireworks
        setup_and_run_fws = self.get_setup_and_run_fw(fw_spec=fw_spec)
        self.setup_fw = setup_and_run_fws['setup_fw']
        self.run_fw = setup_and_run_fws['run_fw']

        # Specify the type of the task that is controlled:
        #  - aborted : the task has been aborted due to a monitoring controller during the Run Task, the FW state
        #              is COMPLETED
        #  - completed : the task has completed, the FW state is COMPLETE
        #  - failed : the task has failed, the FW state is FIZZLED
        if self.run_fw.state == 'COMPLETED':
            if 'src_run_task_aborted' in fw_spec:
                self.control_procedure.set_controlled_item_type(ControlledItemType.task_aborted())
            else:
                self.control_procedure.set_controlled_item_type(ControlledItemType.task_completed())
        elif self.run_fw.state == 'FIZZLED':
            self.control_procedure.set_controlled_item_type(ControlledItemType.task_failed())
        else:
            raise RuntimeError('The state of the Run Firework is "{}" '
                               'while it should be COMPLETED or FIZZLED'.format(self.run_fw.state))

        # Get the keyword_arguments to be passed to the process method of the control_procedure
        #TODO: how to do that kind of automatically ??
        # each key should have : how to get it from the run_fw/(setup_fw)
        #                        how to force/apply it to the next SRC (e.g. how do we say to setup that)
        # Actually, the object, can come from : the setup_fw or from the run_fw (from the setup_spec, from the run_spec,
        # from the setup_task or the run_task (or in general one of the tasks ...
        # even multiple tasks is not yet supported ... should it be ? or should we stay with only one task allways ?)
        # If it is modified, it should update the corresponding bit (setup_spec and/or run_spec and/or
        # setup_task and/or run_task)
        initial_objects_info = self.get_initial_objects_info(setup_fw=self.setup_fw, run_fw= self.run_fw,
                                                             src_directories=self.src_directories)
        qerr_filepath = os.path.join(self.run_fw.launches[-1].launch_dir, 'queue.qerr')
        qout_filepath = os.path.join(self.run_fw.launches[-1].launch_dir, 'queue.qout')
        initial_objects_info.update({'queue_adapter': {'object': self.run_fw.spec['qtk_queueadapter'],
                                                       'updates': [{'target': 'fw_spec',
                                                                    'key': 'qtk_queueadapter'},
                                                                   {'target': 'fw_spec',
                                                                    'key': '_queueadapter',
                                                                    'mod': 'get_subs_dict'}]},
                                     'qerr_filepath': {'object': qerr_filepath},
                                     'qout_filepath': {'object': qout_filepath}})
        initial_objects = {name: obj_info['object'] for name, obj_info in initial_objects_info.items()}
        control_report = self.control_procedure.process(**initial_objects)

        if control_report.unrecoverable:
            f = open(os.path.join(self.control_dir, 'control_report.json'), 'w')
            json.dump(control_report.as_dict(), f)
            f.close()
            #TODO: apply the cleaning here
            if self.src_cleaning is not None:
                pass
            raise ValueError('Errors are unrecoverable. Control report written in "control_report.json"')

        # If everything is ok, update the spec of the children
        if control_report.finalized:
            stored_data = {'control_report': control_report, 'finalized': True}
            update_spec = {}
            mod_spec = []
            run_task = self.run_fw.tasks[-1]
            setup_task = self.setup_fw.tasks[-1]
            task_type = run_task.task_type
            #TODO: should we also add here the cluster in which the calculation was performed so that if the next
            #      SRC trio starts on another cluster, it should fetch the needed files from the run_dir of this cluster
            task_info = {'dir': self.run_dir}
            task_info.update(run_task.additional_task_info())
            task_info.update(setup_task.additional_task_info())
            mod_spec.append({'_push': {'previous_fws->'+task_type: task_info}})
            if self.src_cleaning is not None:
                pass
            return FWAction(stored_data=stored_data, exit=False, update_spec=update_spec, mod_spec=mod_spec,
                            additions=None, detours=None, defuse_children=False)

        # Check the maximum number of restarts
        if task_index.index == self.max_restarts:
            # TODO: decide when to apply cleaning here ?
            if self.src_cleaning is not None:
                pass
            raise ValueError('Maximum number of restarts ({:d}) reached'.format(self.max_restarts))

        # Increase the task_index
        task_index.increase_index()

        # Apply the actions on the objects to get the modified objects (to be passed to SetupTask)
        # modified_objects = {}
        # for target, action in control_report.actions.items():
        #     # Special case right now for the queue adapter ...
        #     if target == 'queue_adapter':
        #         qtk_qadapter = initial_objects[target]
        #         action.apply(qtk_qadapter)
        #         modified_objects['qtk_queueadapter'] = qtk_qadapter
        #         modified_objects['_queueadapter'] = qtk_qadapter.get_subs_dict()
        #     else:
        #         modified_objects[target] = action.apply(initial_objects[target])

        # New spec
        # remove "_tasks" which is present in spec for recent fireworks versions. Remove it here to avoid
        # problems with deepcopy.
        new_spec = dict(self.run_fw.spec)
        new_spec.pop("_tasks", None)
        new_spec = copy.deepcopy(new_spec)

        # New tasks
        setup_task = self.setup_fw.tasks[-1]
        run_task = self.run_fw.tasks[-1]
        control_task = self

        if 'src_modified_objects' in fw_spec:
            modified_objects = fw_spec['src_modified_objects']
        else:
            modified_objects = {}
        setup_spec_update = {}
        run_spec_update = {}
        for target, action in control_report.actions.items():
            target_object = initial_objects[target]
            action.apply(target_object)
            if target not in initial_objects_info:
                raise ValueError('Object "{}" to be modified was not in the initial_objects'.format(target))
            if 'updates' not in initial_objects_info[target]:
                raise ValueError('Update information not present for object "{}"'.format(target))
            for update in initial_objects_info[target]['updates']:
                if update['target'] == 'fw_spec':
                    if 'mod' in update:
                        mod = getattr(target_object, update['mod'])()
                        new_spec[update['key']] = mod
                        modified_objects[update['key']] = mod
                    else:
                        new_spec[update['key']] = target_object
                        modified_objects[update['key']] = target_object
                elif update['target'] == 'setup_fw_spec':
                    if 'mod' in update:
                        mod = getattr(target_object, update['mod'])()
                        setup_spec_update[update['key']] = mod
                    else:
                        setup_spec_update[update['key']] = target_object
                elif update['target'] == 'run_fw_spec':
                    if 'mod' in update:
                        mod = getattr(target_object, update['mod'])()
                        run_spec_update[update['key']] = mod
                    else:
                        run_spec_update[update['key']] = target_object
                elif update['target'] in ['setup_task', 'run_task']:
                    task = setup_task if update['target'] == 'setup_task' else run_task
                    attr = getattr(task, update['attribute'])
                    if 'mod' in update:
                        mod = getattr(target_object, update['mod'])()
                        attr = mod
                    else:
                        attr = target_object
                elif 'setup_task' in update['target']:
                    sp = update['target'].split('.')
                    if len(sp) != 2:
                        raise ValueError('target is "{}" and contains more than 1 "."'.format(update['target']))
                    if sp[0] != 'setup_task':
                        raise ValueError('target does not start with "setup_task" ...')
                    task = setup_task
                    task_attribute = getattr(task, sp[1])
                    if 'attribute' in update:
                        attr = getattr(task_attribute, update['attribute'])
                        if 'mod' in update:
                            mod = getattr(target_object, update['mod'])()
                            attr = mod
                        else:
                            attr = target_object
                    elif 'setter' in update:
                        setter = getattr(task_attribute, update['setter'])
                        if 'mod' in update:
                            mod = getattr(target_object, update['mod'])()
                            setter(mod)
                        else:
                            setter(target_object)
                else:
                    raise ValueError('Only changes to fw_spec, setup_task and run_task are allowed right now ...')

        # Set the restart_info
        setup_task.set_restart_info(control_report.restart_info)

        # Pass the modified objects to the next SetupTask
        new_spec['src_modified_objects'] = modified_objects
        new_spec.pop('_launch_dir')
        new_spec.pop('src_directories')
        new_spec['previous_src'] = {'src_directories': self.src_directories}
        if 'all_src_directories' in new_spec:
            new_spec['all_src_directories'].append({'src_directories': self.src_directories})
        else:
            new_spec['all_src_directories'] = [{'src_directories': self.src_directories}]
        # if '_queueadapter' in modified_objects:
        #     new_spec['_queueadapter'] = modified_objects['_queueadapter']
        #TODO: what to do here ? Right now this should work, just transfer information from the run_fw to the
        # next SRC group
        if 'previous_fws' in fw_spec:
            new_spec['previous_fws'] = fw_spec['previous_fws']
        # Create the new SRC trio
        # TODO: check initialization info, deps, ... previous_fws, ... src_previous_fws ? ...
        new_SRC_fws = createSRCFireworks(setup_task=setup_task, run_task=run_task, control_task=control_task,
                                         spec=new_spec, initialization_info=None, task_index=task_index,
                                         run_spec_update=run_spec_update, setup_spec_update=setup_spec_update)
        wf = Workflow(fireworks=new_SRC_fws['fws'], links_dict=new_SRC_fws['links_dict'])
        return FWAction(stored_data={'control_report': control_report}, detours=[wf])

    def get_setup_and_run_fw(self, fw_spec):
        # Get the launchpad
        if '_add_launchpad_and_fw_id' in fw_spec:
            lp = self.launchpad
            control_fw_id = self.fw_id
        else:
            try:
                fw_dict = loadfn('FW.json')
            except IOError:
                try:
                    fw_dict = loadfn('FW.yaml')
                except IOError:
                    raise RuntimeError("Launchpad/fw_id not present in spec and No FW.json nor FW.yaml file present: "
                                       "impossible to determine fw_id")
            lp = LaunchPad.auto_load()
            control_fw_id = fw_dict['fw_id']
        # Check that this ControlTask has only one parent firework
        this_lzy_wf = lp.get_wf_by_fw_id_lzyfw(control_fw_id)
        parents_fw_ids = this_lzy_wf.links.parent_links[control_fw_id]
        if len(parents_fw_ids) != 1:
            raise ValueError('ControlTask\'s Firework should have exactly one parent firework')
        run_fw_id = parents_fw_ids[0]
        # Get the Run Firework and its state
        run_fw = lp.get_fw_by_id(fw_id=run_fw_id)
        run_is_fizzled = '_fizzled_parents' in fw_spec
        if run_is_fizzled and not run_fw.state == 'FIZZLED':
            raise ValueError('ControlTask has "_fizzled_parents" key but parent Run firework is not fizzled ...')
        run_is_completed = run_fw.state == 'COMPLETED'
        if run_is_completed and run_is_fizzled:
            raise ValueError('Run firework is FIZZLED and COMPLETED ...')
        if (not run_is_completed) and (not run_is_fizzled):
            raise ValueError('Run firework is neither FIZZLED nor COMPLETED ...')
        # Get the Setup Firework
        setup_job_info = run_fw.spec['_job_info'][-1]
        setup_fw_id = setup_job_info['fw_id']
        setup_fw = lp.get_fw_by_id(fw_id=setup_fw_id)
        return {'setup_fw': setup_fw, 'run_fw': run_fw}

    def get_initial_objects_info(self, setup_fw, run_fw, src_directories):
        return {}

    @classmethod
    def from_controllers(cls, controllers, max_restarts=10):
        cp = ControlProcedure(controllers=controllers)
        return cls(control_procedure=cp, max_restarts=max_restarts)

    @serialize_fw
    def to_dict(self):
        return {'control_procedure': self.control_procedure.as_dict(),
                'manager': self.manager.as_dict() if self.manager is not None else None,
                'max_restarts': self.max_restarts,
                'src_cleaning': self.src_cleaning.as_dict() if self.src_cleaning is not None else None}

    @classmethod
    def from_dict(cls, d):
        control_procedure = ControlProcedure.from_dict(d['control_procedure'])
        dec = MontyDecoder()
        if d['manager'] is None:
            manager = None
        else:
            manager = dec.process_decoded(d['manager']),
        if 'src_cleaning' in d:
            src_cleaning = SRCCleaning.from_dict(d['src_cleaning']) if d['src_cleaning'] is not None else None
        else:
            src_cleaning = None
        return cls(control_procedure=control_procedure, manager=manager, max_restarts=d['max_restarts'],
                   src_cleaning=src_cleaning)


class SRCCleanerOptions(MSONable):

    WHEN_TO_CLEAN = ['EACH_STEP', 'LAST_STEP', 'EACH_STEP_EXCEPT_LAST']
    CURRENT_SRC_STATES_ALLOWED = ['RECOVERABLE', 'UNRECOVERABLE', 'MAXRESTARTS', 'FINALIZED']

    def __init__(self, when_to_clean, current_src_states_allowed, which_src_steps_to_clean):
        self.when_to_clean = when_to_clean
        self.current_src_states_allowed = current_src_states_allowed
        self.which_src_steps_to_clean = which_src_steps_to_clean

    @classmethod
    def clean_all(cls):
        return cls(when_to_clean='EACH_STEP', current_src_states_allowed=cls.CURRENT_SRC_STATES_ALLOWED,
                   which_src_steps_to_clean='all')

    @classmethod
    def clean_all_except_last(cls):
        return cls(when_to_clean='EACH_STEP', current_src_states_allowed=cls.CURRENT_SRC_STATES_ALLOWED,
                   which_src_steps_to_clean='all_before_this_one')

    @property
    def when_to_clean(self):
        return self._when_to_clean

    @when_to_clean.setter
    def when_to_clean(self, when_to_clean):
        if when_to_clean not in self.WHEN_TO_CLEAN:
            raise ValueError('Argument "when_to_clean" is "{}" while it should be one of the following : '
                             '{}'.format(when_to_clean,
                                         ', '.join(self.WHEN_TO_CLEAN)))
        self._when_to_clean = when_to_clean

    @property
    def current_src_states_allowed(self):
        return self._current_src_states_allowed

    @current_src_states_allowed.setter
    def current_src_states_allowed(self, current_src_states_allowed):
        for current_src_state_allowed in current_src_states_allowed:
            if current_src_state_allowed not in self.CURRENT_SRC_STATES_ALLOWED:
                raise ValueError('One of the items in  "current_src_states_allowed" is "{}" while it should be one of '
                                 'the following : {}'.format(current_src_state_allowed,
                                                             ', '.join(self.CURRENT_SRC_STATES_ALLOWED)))
        self._current_src_states_allowed = current_src_states_allowed

    @property
    def which_src_steps_to_clean(self):
        return self._which_src_steps_to_clean

    @which_src_steps_to_clean.setter
    def which_src_steps_to_clean(self, which_src_steps_to_clean):
        if which_src_steps_to_clean in ['all', 'this_one', 'all_before_this_one', 'all_before_the_previous_one',
                                        'the_one_before_this_one', 'the_one_before_the_previous_one']:
            self._which_src_steps_to_clean = which_src_steps_to_clean
            self._which_src_steps_to_clean_pattern = which_src_steps_to_clean
        elif which_src_steps_to_clean[:7] == 'single_':
            sp = which_src_steps_to_clean.split('_')
            if len(sp) != 2:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "single_" but has more '
                                 'than one underscore ... '
                                 'Impossible to identify the step to clean.'.format(which_src_steps_to_clean))
            try:
                istep = int(sp[1])
            except ValueError:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "single_" but the '
                                 'remaining part is not an integer ... '
                                 'Impossible to identify the step to clean.'.format(which_src_steps_to_clean))
            if istep < 1:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "single_" but the '
                                 'remaining part is an integer < 1 ... '
                                 'Impossible to identify the step to clean.'.format(which_src_steps_to_clean))
            self._which_src_steps_to_clean = which_src_steps_to_clean
            self._which_src_steps_to_clean_pattern = 'single_N'
        elif (len(which_src_steps_to_clean) > 29 and which_src_steps_to_clean[:15] == 'all_before_the_' and
                  which_src_steps_to_clean[-14:] == '_previous_ones'):
            sp = which_src_steps_to_clean.split('_')
            if len(sp) != 6:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "all_before_the_", '
                                 'ends with "_previous_ones" but has more than 5 underscores ... Impossible to '
                                 'identify the steps to clean.'.format(which_src_steps_to_clean))
            try:
                istep = int(sp[3])
            except ValueError:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "all_before_the_", '
                                 'ends with "_previous_ones" but the remaining part is not an integer ... '
                                 'Impossible to identify the steps to clean.'.format(which_src_steps_to_clean))
            if istep < 2:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "all_before_the_", '
                                 'ends with "_previous_ones" but the remaining part an integer less than 2 ... '
                                 'Impossible to identify the steps to clean.'.format(which_src_steps_to_clean))
            self._which_src_steps_to_clean = which_src_steps_to_clean
            self._which_src_steps_to_clean_pattern = 'all_before_the_N_previous_ones'
        elif (len(which_src_steps_to_clean) > 33 and which_src_steps_to_clean[:19] == 'the_one_before_the_' and
                  which_src_steps_to_clean[-14:] == '_previous_ones'):
            sp = which_src_steps_to_clean.split('_')
            if len(sp) != 7:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "the_one_before_the_", '
                                 'ends with "_previous_ones" but has more than 6 underscores ... Impossible to '
                                 'identify the steps to clean.'.format(which_src_steps_to_clean))
            try:
                istep = int(sp[4])
            except ValueError:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "the_one_before_the_", '
                                 'ends with "_previous_ones" but the remaining part is not an integer ... '
                                 'Impossible to identify the steps to clean.'.format(which_src_steps_to_clean))
            if istep < 2:
                raise ValueError('Argument "which_src_steps_to_clean" is "{}". It starts with "the_one_before_the_", '
                                 'ends with "_previous_ones" but the remaining part an integer less than 2 ... '
                                 'Impossible to identify the steps to clean.'.format(which_src_steps_to_clean))
            self._which_src_steps_to_clean = which_src_steps_to_clean
            self._which_src_steps_to_clean_pattern = 'the_one_before_the_N_previous_ones'
        #TODO: implement "the_M_before_the_N_previous_ones" if needed ...
        else:
            raise ValueError('Argument "which_src_steps_to_clean" is "{}". This is not allowed. See documentation for '
                             'the allowed options.'.format(which_src_steps_to_clean))

    def steps_to_clean(self, this_step_index, this_step_state):
        if this_step_state not in self.current_src_states_allowed:
            return []
        if self._which_src_steps_to_clean_pattern == 'all':
            return list(range(1, this_step_index+1))
        elif self._which_src_steps_to_clean_pattern == 'this_one':
            return [this_step_index]
        elif self._which_src_steps_to_clean_pattern == 'the_one_before_this_one':
            if this_step_index == 1:
                return []
            return [this_step_index-1]
        elif self._which_src_steps_to_clean_pattern == 'the_one_before_the_previous_one':
            if this_step_index <= 2:
                return []
            return [this_step_index-2]
        elif self._which_src_steps_to_clean_pattern == 'the_one_before_the_N_previous_ones':
            iprev = int(self.which_src_steps_to_clean.split('_')[4])
            istep = this_step_index-iprev-1
            if istep < 1:
                return []
            return [istep]
        elif self._which_src_steps_to_clean_pattern == 'all_before_this_one':
            return list(range(1, this_step_index))
        elif self._which_src_steps_to_clean_pattern == 'all_before_the_previous_one':
            return list(range(1, this_step_index-1))
        elif self._which_src_steps_to_clean_pattern == 'all_before_the_N_previous_ones':
            iprev = int(self.which_src_steps_to_clean.split('_')[3])
            return list(range(1, this_step_index-iprev))
        elif self._which_src_steps_to_clean_pattern == 'single_N':
            istep = int(self.which_src_steps_to_clean.split('_')[1])
            if istep > this_step_index:
                return []
            return [istep]
        raise ValueError('Should not reach this point in "steps_to_clean" of "SRCCleanerOptions"')

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'when_to_clean': self.when_to_clean,
                'current_src_states_allowed': self.current_src_states_allowed,
                'which_src_steps_to_clean': self.which_src_steps_to_clean}

    @classmethod
    def from_dict(cls, d):
        return cls(when_to_clean=d['when_to_clean'],
                   current_src_states_allowed=d['current_src_states_allowed'],
                   which_src_steps_to_clean=d['which_src_steps_to_clean'])


class SRCCleaner(MSONable):

    # RECURRENCE_TYPES = ['END', 'STEP', 'STEP_EXCEPT_END']
    # STATE_TYPES = {'END': ['FINALIZED', 'UNRECOVERABLE', 'MAXRESTARTS'],
    #                'STEP': ['RECOVERABLE', 'UNRECOVERABLE', 'MAXRESTARTS', 'FINALIZED'],
    #                'STEP_EXCEPT_END': ['RECOVERABLE']}
    # CLEANING_TYPES = ['ALL', 'ALL_EXCEPT_LAST', 'LAST']
    SRC_TYPES = ['setup', 'run', 'control', 'src_root']

    def __init__(self, cleaners=None, src_type='run', cleaner_options=SRCCleanerOptions.clean_all()):
        if cleaners is None:
            self.cleaners = []
        else:
            self.cleaners = cleaners
        self.src_type = src_type
        self.cleaner_options = cleaner_options

    @property
    def cleaners(self):
        return self._cleaners

    @cleaners.setter
    def cleaners(self, cleaners):
        if cleaners is None:
            self._cleaners = []
        elif isinstance(cleaners, list):
            for cl in cleaners:
                if not isinstance(cl, Cleaner):
                    raise ValueError('One of the items in cleaners is not a Cleaner instance but is an instance '
                                     'of {}'.format(cl.__class__.__name__))
            self._cleaners = cleaners
        else:
            raise ValueError('The variable "cleaners" should be either None or a list of Cleaner objects')

    @property
    def src_type(self):
        return self._src_type

    @src_type.setter
    def src_type(self, src_type):
        if src_type not in self.SRC_TYPES:
            raise ValueError('Argument "src_type" should be one of the following : '
                             '{}'.format(', '.format(self.SRC_TYPES)))
        self._src_type = src_type

    def src_dir_to_clean(self, src_directories):
        return src_directories['{}_dir'.format(self.src_type)]

    def check_recurrence(self, src_task_index, state):
        if state == self.recurrence:
            return True
        return False

    def clean(self, last_src_directories, previous_src_dirs, src_task_index, state):
        dirs_to_clean = []
        if self.cleaning in ['ALL', 'LAST']:
            dirs_to_clean.append(self.src_dir_to_clean(last_src_directories))
            pass
        if self.cleaning in ['ALL', 'ALL_EXCEPT_LAST']:
            for src_directories in previous_src_dirs:
                dirs_to_clean.append(self.src_dir_to_clean(src_directories))
        for dir_to_clean in dirs_to_clean:
            for cleaner in self.cleaners:
                cleaner.clean(root_directory=dir_to_clean)

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'cleaners': [c.as_dict() for c in self.cleaners],
                'src_types': self.src_type,
                'cleaner_options': self.cleaner_options}

    @classmethod
    def from_dict(cls, d):
        return cls(recurrence=d['recurrence'],
                   cleaners=[Cleaner.from_dict(d_c) for d_c in d['cleaners']],
                   src_type=d['src_type'])


class SRCCleaning(MSONable):

    def __init__(self, src_cleaners=None):
        if src_cleaners is None:
            self.src_cleaners = []
        else:
            self.src_cleaners = src_cleaners

    def clean(self, src_directories, previous_src_dirs, state):
        pass

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'src_cleaners': [src_c.as_dict() for src_c in self.src_cleaners]}

    @classmethod
    def from_dict(cls, d):
        return cls(src_cleaners=[SRCCleaner.from_dict(d_src_c) for d_src_c in d['src_cleaners']])


def createSRCFireworks(setup_task, run_task, control_task, spec=None, initialization_info=None,
                       task_index=None, setup_spec_update=None, run_spec_update=None):
    # Make a full copy of the spec
    if spec is None:
        spec = {}
    if initialization_info is None:
        initialization_info = {}
    spec = copy.deepcopy(spec)
    spec['_add_launchpad_and_fw_id'] = True
    spec['_add_fworker'] = True
    # Initialize the SRC task_index
    if task_index is not None:
        src_task_index = SRCTaskIndex.from_any(task_index)
    else:
        # src_task_index = SRCTaskIndex.from_any('unknown-task')
        src_task_index = SRCTaskIndex.from_task(run_task)
    spec['SRC_task_index'] = src_task_index

    # SetupTask
    setup_spec = copy.deepcopy(spec)
    # Remove any initial queue_adapter_update from the spec
    setup_spec.pop('queue_adapter_update', None)

    setup_spec = set_short_single_core_to_spec(setup_spec)
    setup_spec['_preserve_fworker'] = True
    setup_spec['_pass_job_info'] = True
    setup_spec['initialization_info'] = initialization_info
    setup_spec.update({} if setup_spec_update is None else setup_spec_update)
    setup_fw = Firework(setup_task, spec=setup_spec, name=src_task_index.setup_str)

    # RunTask
    run_spec = copy.deepcopy(spec)
    run_spec['SRC_task_index'] = src_task_index
    run_spec['_preserve_fworker'] = True
    run_spec['_pass_job_info'] = True
    run_spec.update({} if run_spec_update is None else run_spec_update)
    run_fw = Firework(run_task, spec=run_spec, name=src_task_index.run_str)

    # ControlTask
    control_spec = copy.deepcopy(spec)
    control_spec = set_short_single_core_to_spec(control_spec)
    control_spec['SRC_task_index'] = src_task_index
    control_spec['_allow_fizzled_parents'] = True
    control_fw = Firework(control_task, spec=control_spec, name=src_task_index.control_str)

    links_dict = {setup_fw.fw_id: [run_fw.fw_id],
                  run_fw.fw_id: [control_fw.fw_id]}
    return {'setup_fw': setup_fw, 'run_fw': run_fw, 'control_fw': control_fw, 'links_dict': links_dict,
            'fws': [setup_fw, run_fw, control_fw]}


class SRCTaskIndex(MSONable):

    ALLOWED_CHARS = ['-']

    def __init__(self, task_type, index=1):
        self.set_task_type(task_type=task_type)
        self.index = index

    def set_task_type(self, task_type):
        prefix_test_string = str(task_type)
        for allowed_char in self.ALLOWED_CHARS:
            prefix_test_string = prefix_test_string.replace(allowed_char, "")
        if not prefix_test_string.isalnum():
            ac_str = ', '.join(['"{}"'.format(ac) for ac in self.ALLOWED_CHARS])
            raise ValueError('task_type should only contain letters '
                             'and the following characters : {}'.format(ac_str))
        self.task_type = task_type

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        if isinstance(index, int):
            self._index = index
        elif isinstance(index, str):
            try:
                myindex = int(index)
                self._index = myindex
            except:
                raise ValueError('Index in SRCTaskIndex should be an integer or a string '
                                 'that can be cast into an integer')
        else:
            raise ValueError('Index in SRCTaskIndex should be an integer or a string '
                             'that can be cast into an integer')

    def increase_index(self):
        self.index += 1

    def __add__(self, other):
        if not isinstance(other, int):
            raise ValueError('The __add__ method in SRCTaskIndex should be an integer')
        self.index += other

    def __str__(self):
        return '_'.join([self.task_type, str(self.index)])

    @property
    def setup_str(self):
        return '_'.join(['setup', self.__str__()])

    @property
    def run_str(self):
        return '_'.join(['run', self.__str__()])

    @property
    def control_str(self):
        return '_'.join(['control', self.__str__()])

    @classmethod
    def from_string(cls, SRC_task_index_string):
        sp = SRC_task_index_string.split('_')
        # if len(sp) not in [2, 3]:
        #     raise ValueError('SRC_task_index_string should contain 1 or 2 underscores ("_") '
        #                      'while it contains {:d}'.format(len(sp)-1))
        if len(sp) == 1:
            return cls(task_type=sp[0])
        if any([len(part) == 0 for part in sp]):
            raise ValueError('SRC_task_index_string has an empty part when separated by underscores ...')
        if len(sp) == 2:
            return cls(task_type=sp[0], index=sp[1])
        elif len(sp) == 3:
            if sp[0] not in ['setup', 'run', 'control']:
                raise ValueError('SRC_task_index_string should start with "setup", "run" or "control" when 3 parts are '
                                 'identified')
            return cls(task_type=sp[1], index=sp[2])

    @classmethod
    def from_any(cls, SRC_task_index):
        if isinstance(SRC_task_index, six.string_types):
            return cls.from_string(SRC_task_index)
        elif isinstance(SRC_task_index, SRCTaskIndex):
            return cls(task_type=SRC_task_index.task_type, index=SRC_task_index.index)
        else:
            raise ValueError('SRC_task_index should be an instance of "str" or "SRCTaskIndex" '
                             'in "from_any" class method')

    @classmethod
    def from_task(cls, task):
        return cls(task_type=task.task_type)

    @classmethod
    def from_dict(cls, d):
        return cls(task_type=d['task_type'], index=d['index'])

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'task_type': self.task_type,
                'index': self.index}



def get_queue_adapter_update(qtk_queueadapter, corrections, qa_params=None):
    if qa_params is None:
        qa_params = ['timelimit', 'mem_per_proc', 'master_mem_overhead']
    queue_adapter_update = {}
    for qa_param in qa_params:
        if qa_param == 'timelimit':
            queue_adapter_update[qa_param] = qtk_queueadapter.timelimit
        elif qa_param == 'mem_per_proc':
            queue_adapter_update[qa_param] = qtk_queueadapter.mem_per_proc
        elif qa_param == 'master_mem_overhead':
            queue_adapter_update[qa_param] = qtk_queueadapter.master_mem_overhead
        else:
            raise ValueError('Wrong queue adapter parameter for update')
    for correction in corrections:
        for action in correction['actions']:
            if action['object']['key'] == 'qtk_queueadapter':
                qa_update = action['action']['_set']
                queue_adapter_update.update(qa_update)
    return queue_adapter_update


################
# Exceptions
################

class SRCError(Exception):
    pass


class SetupError(SRCError):
    pass


class RunError(SRCError):
    pass


class ControlError(SRCError):
    pass


################
# Timings
################


class FWTime(MSONable):
    def __init__(self, fw_name, fw_id, ncpus, fwtime_secs, clustertime_secs=None):
        self.fw_name = fw_name
        self.fw_id = fw_id
        self.ncpus = ncpus
        self.fwtime_secs = fwtime_secs
        self.clustertime_secs = clustertime_secs

    @property
    def time_per_cpu(self):
        if self.clustertime_secs is not None:
            return self.clustertime_secs
        return self.fwtime_secs

    @property
    def total_time(self):
        return self.ncpus*self.time_per_cpu

    def as_dict(self):
        dd = dict(fw_name=self.fw_name, fw_id=self.fw_id,
                  ncpus=self.ncpus, fwtime_secs=self.fwtime_secs,
                  clustertime_secs=self.clustertime_secs)
        return dd

    @classmethod
    def from_dict(cls, d):
        return cls(fw_name=d['fw_name'], fw_id=d['fw_id'],
                   ncpus=d['ncpus'], fwtime_secs=d['fwtime_secs'], clustertime_secs=d['clustertime_secs'])

    @classmethod
    def from_fw_id(cls, fw_id, lpad=None):
        if lpad is None:
            lpad = LaunchPad.auto_load()
        fw_dict = lpad.get_fw_dict_by_id(fw_id=fw_id)
        name = fw_dict['name']
        # TODO: find a way to know the number of cpus here ? Or should we always assume it is 1 ?
        ncpus = 1
        fwtime_secs = 0.0
        # TODO: get the runtime from the cluster (taking the reservation_id and things like that ?)
        clustertime_secs = None
        return cls(fw_name=name, fw_id=fw_id,
                   ncpus=ncpus, fwtime_secs=fwtime_secs, clustertime_secs=clustertime_secs)

class SRCFWTime(FWTime):
    def __init__(self, fw_name, fw_id, ncpus, fwtime_secs, clustertime_secs=None,
                 src_type=None, task_type=None, task_index=None):
        super(SRCFWTime, self).__init__(fw_name=fw_name, fw_id=fw_id, ncpus=ncpus,
                                        fwtime_secs=fwtime_secs, clustertime_secs=clustertime_secs)
        self.src_type = src_type
        self.task_type = task_type
        self.task_index = task_index

    def as_dict(self):
        dd = dict(src_type=self.src_type, task_type=self.task_type, task_index=self.task_index.as_dict(),
                  ncpus=self.ncpus, fwtime_secs=self.fwtime_secs,
                  clustertime_secs=self.clustertime_secs)
        return dd

    @classmethod
    def from_dict(cls, d):
        return cls(src_type=d['src_type'], task_type=d['task_type'], task_index=SRCTaskIndex.from_any(d['task_index']),
                   ncpus=d['ncpus'], fwtime_secs=d['fwtime_secs'], clustertime_secs=d['clustertime_secs'])

    @classmethod
    def from_fw_id(cls, fw_id, lpad=None):
        if lpad is None:
            lpad = LaunchPad.auto_load()
        fw_dict = lpad.get_fw_dict_by_id(fw_id=fw_id)
        name = fw_dict['name']
        # TODO: find a way to know the number of cpus here ? Or should we always assume it is 1 ?
        ncpus = 1
        fwtime_secs = 0.0


# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

from abiflows.core.testing import AbiflowsTest
from abiflows.utils.factors import lowest_nn_gte_mm

class TestFactors(AbiflowsTest):
    def test_lowest_nn_gte_mm(self):
        ll, exponents = lowest_nn_gte_mm(101, [2, 3, 5])
        self.assertEqual(ll, 108)
        self.assertEqual(exponents, (2, 3, 0))

        ll, exponents = lowest_nn_gte_mm(100, [2, 3, 5])
        self.assertEqual(ll, 100)
        self.assertEqual(exponents, (2, 0, 2))

        ll, exponents = lowest_nn_gte_mm(100, [2, 3, 5])
        self.assertEqual(ll, 100)
        self.assertEqual(exponents, (2, 0, 2))

        ll, exponents = lowest_nn_gte_mm(101322142376, [7, 2, 13])
        self.assertEqual(ll, 101322142376)
        self.assertEqual(exponents, (8, 3, 3))

        ll, exponents = lowest_nn_gte_mm(101322142375, [7, 2, 13])
        self.assertEqual(ll, 101322142376)
        self.assertEqual(exponents, (8, 3, 3))

        ll, exponents = lowest_nn_gte_mm(101322142377, [7, 2, 13])
        self.assertEqual(ll, 102820990636)
        self.assertEqual(exponents, (11, 2, 1))


from __future__ import print_function, division, unicode_literals, absolute_import

import os
import pytest
import abipy.abilab as abilab
import abiflows.fireworks.tasks.abinit_common
import abiflows.fireworks.tasks.abinit_tasks as abinit_tasks

from abipy.electrons.gsr import GsrFile
from abipy.flowtk.utils import Directory
from fireworks.core.rocket_launcher import rapidfire
from abiflows.fireworks.workflows.abinit_workflows import InputFWWorkflow
from abiflows.fireworks.tasks.abinit_tasks import ScfFWTask

ABINIT_VERSION = "7.11.5"

# pytestmark = [pytest.mark.skipif(not has_abinit(ABINIT_VERSION), reason="Abinit version {} is not in PATH".format(ABINIT_VERSION)),
#               pytest.mark.skipif(not has_fireworks(), reason="fireworks paackage is missing"),
#               pytest.mark.skipif(not has_mongodb(), reason="no connection to mongodb")]

pytestmark = [pytest.mark.usefixtures("cleandb"), pytest.mark.check_abiflow]


def match_results(fw, abitask):
    fw_gsr_path = Directory(os.path.join(fw.launches[-1].launch_dir, abiflows.fireworks.tasks.abinit_common.OUTDIR_NAME)).has_abiext("GSR")
    with GsrFile(fw_gsr_path) as gsr1, abitask.open_gsr() as gsr2:
        if gsr1.energy - gsr2.energy > 0.0001:
            return False

    return True


class ItestCheckScf():

    def itest_scf(self, lp, fworker, fwp, tmpdir, benchmark_input_scf):
        wf = InputFWWorkflow(benchmark_input_scf, task_type=ScfFWTask)

        scf_fw_id = wf.fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        fw = lp.get_fw_by_id(scf_fw_id)

        assert fw.state == "COMPLETED"

        # Build the flow
        flow = abilab.Flow(fwp.workdir, manager=fwp.manager)
        work = flow.register_task(benchmark_input_scf, task_class=abilab.ScfTask)

        flow.allocate()
        flow.build_and_pickle_dump()

        # Run t0, and check status
        t0 = work[0]
        t0.start_and_wait()
        t0.check_status()
        assert t0.status == t0.S_OK

        assert match_results(fw, t0)

    def itest_scf_not_converged(self, lp, fworker, fwp, tmpdir, benchmark_input_scf):
        old_cwd = os.getcwd()
        benchmark_input_scf.set_vars(nstep=4)

        wf = InputFWWorkflow(benchmark_input_scf, task_type=ScfFWTask)

        scf_fw_id = wf.fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        while lp.run_exists(fworker):
            rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        num_restarts_fw = wf.fws[-1].tasks[0].restart_info.num_restarts

        # Build the flow
        flow = abilab.Flow(fwp.workdir, manager=fwp.manager)
        work = flow.register_task(benchmark_input_scf, task_class=abilab.ScfTask)

        flow.allocate()
        flow.build_and_pickle_dump()

        # go to the main dir (to have the abipy configuration files)
        os.chdir(old_cwd)

        # Run t0, and check status
        t0 = work[0]
        assert flow.make_scheduler().start() == 0

        num_restarts_abiflow = t0.num_restarts

        assert num_restarts_fw == num_restarts_abiflow

        assert match_results(lp.get_wf_by_fw_id(scf_fw_id).fws[-1], t0)












from __future__ import print_function, division, unicode_literals, absolute_import

import pytest
import os
import glob
import numpy as np

from fireworks.core.rocket_launcher import rapidfire
from abiflows.fireworks.workflows.abinit_workflows import NscfFWWorkflow
from abiflows.fireworks.tasks.abinit_tasks import NscfFWTask
from abiflows.fireworks.utils.fw_utils import load_abitask, get_fw_by_task_index
from abiflows.core.testing import AbiflowsIntegrationTest

ABINIT_VERSION = "8.6.1"

# pytestmark = [pytest.mark.skipif(not has_abinit(ABINIT_VERSION), reason="Abinit version {} is not in PATH".format(ABINIT_VERSION)),
#               pytest.mark.skipif(not has_fireworks(), reason="fireworks paackage is missing"),
#               pytest.mark.skipif(not has_mongodb(), reason="no connection to mongodb")]

pytestmark = pytest.mark.usefixtures("cleandb")

class ItestNscf(AbiflowsIntegrationTest):

    def itest_nscf_wf(self, lp, fworker, tmpdir, input_ebands_si_low, use_autoparal):
        """
        Simple test of NscfFWWorkflow with autoparal True and False.
        """
        wf = NscfFWWorkflow(*input_ebands_si_low, autoparal=False)

        wf.add_final_cleanup(["WFK"])

        scf_fw_id = wf.scf_fw.fw_id
        nscf_fw_id = wf.nscf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]
        nscf_fw_id = old_new[nscf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        # check the effect of the final cleanup
        scf_task = load_abitask(get_fw_by_task_index(wf, "scf", index=1))

        assert len(glob.glob(os.path.join(scf_task.outdir.path, "*_WFK"))) == 0
        assert len(glob.glob(os.path.join(scf_task.outdir.path, "*_DEN"))) == 1
        assert len(glob.glob(os.path.join(scf_task.tmpdir.path, "*"))) == 0
        assert len(glob.glob(os.path.join(scf_task.indir.path, "*"))) == 0

        if self.check_numerical_values:
            scf_task = load_abitask(lp.get_fw_by_id(scf_fw_id))

            with scf_task.open_gsr() as scf_gsr:
                assert scf_gsr.energy == pytest.approx(-241.239839134, rel=0.01)

            nscf_task = load_abitask(lp.get_fw_by_id(nscf_fw_id))
            with nscf_task.open_gsr() as nscf_gsr:
                assert np.allclose((-6.2581504, 5.5974646, 5.5974646), nscf_gsr.ebands.eigens[0, 0, :3], rtol=0.1)


    def itest_not_converged(self, lp, fworker, tmpdir, input_ebands_si_low):
        """
        Tests the running of the NscfFWWorkflow with non convergence and restart
        """
        input_ebands_si_low[1].set_vars(nstep=5)

        wf = NscfFWWorkflow(*input_ebands_si_low, autoparal=False)

        scf_fw_id = wf.scf_fw.fw_id
        nscf_fw_id = wf.nscf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]
        nscf_fw_id = old_new[nscf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir), nlaunches=2)

        nscf_fw = lp.get_fw_by_id(nscf_fw_id)

        assert nscf_fw.state == "COMPLETED"

        launch = nscf_fw.launches[-1]

        assert any(event.yaml_tag == NscfFWTask.CRITICAL_EVENTS[0].yaml_tag for event in launch.action.stored_data['report'])

        links = lp.get_wf_by_fw_id(scf_fw_id).links

        assert nscf_fw_id in links and len(links[nscf_fw_id]) == 1

        fw_child_id = links[nscf_fw_id][0]
        fw_child = lp.get_fw_by_id(fw_child_id)

        assert fw_child.state == "READY"

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        fw_child = lp.get_fw_by_id(fw_child_id)

        assert fw_child.state == "COMPLETED"

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        if self.check_numerical_values:
            scf_task = load_abitask(lp.get_fw_by_id(scf_fw_id))
            with scf_task.open_gsr() as scf_gsr:
                assert scf_gsr.energy == pytest.approx(-241.239839134, rel=0.01)

            last_nscf_task = load_abitask(get_fw_by_task_index(wf, "nscf", index=-1))
            with last_nscf_task.open_gsr() as nscf_gsr:
                assert np.allclose((-6.2581504, 5.5974646, 5.5974646), nscf_gsr.ebands.eigens[0, 0, :3], rtol=0.1)






# coding: utf-8
"""
Utility tasks for Fireworks.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import copy
import os
import shutil
import logging
import traceback
import importlib
import json

from monty.json import jsanitize
from monty.json import MontyDecoder
from custodian.ansible.interpreter import Modder
from fireworks.core.firework import Firework, FireTaskBase, FWAction, Workflow
from fireworks.core.launchpad import LaunchPad
from fireworks.utilities.fw_utilities import explicit_serialize
from fireworks.utilities.fw_serializers import serialize_fw
from abiflows.fireworks.tasks.abinit_common import TMPDIR_NAME, OUTDIR_NAME, INDIR_NAME
from abiflows.fireworks.utils.custodian_utils import SRCErrorHandler
from abiflows.fireworks.utils.fw_utils import set_short_single_core_to_spec, FWTaskManager, get_lp_and_fw_id_from_task
from abiflows.database.mongoengine.utils import DatabaseData


logger = logging.getLogger(__name__)


SRC_TIMELIMIT_BUFFER = 120


#TODO: make it possible to use "any" task and in particular, MergeDdbTask, AnaDdbTask, ...
# within the SRC scheme (to be rationalized), as well as Task not related to abinit ...
def createSRCFireworksOld(task_class, task_input, SRC_spec, initialization_info, wf_task_index_prefix, current_task_index=1,
                          handlers=None, validators=None,
                          deps=None, task_type=None, queue_adapter_update=None):
    SRC_spec = copy.deepcopy(SRC_spec)
    SRC_spec['initialization_info'] = initialization_info
    SRC_spec['_add_launchpad_and_fw_id'] = True
    SRC_spec['SRCScheme'] = True
    prefix_allowed_chars = ['-']
    prefix_test_string = str(wf_task_index_prefix)
    for allowed_char in prefix_allowed_chars:
        prefix_test_string = prefix_test_string.replace(allowed_char, "")
    if not prefix_test_string.isalnum():
        raise ValueError('wf_task_index_prefix should only contain letters '
                         'and the following characters : {}'.format(prefix_test_string))
    SRC_spec['wf_task_index_prefix'] = wf_task_index_prefix

    # Remove any initial queue_adapter_update from the spec
    SRC_spec.pop('queue_adapter_update', None)
    if queue_adapter_update is not None:
        SRC_spec['queue_adapter_update'] = queue_adapter_update

    # Setup (Autoparal) run
    SRC_spec_setup = copy.deepcopy(SRC_spec)
    SRC_spec_setup = set_short_single_core_to_spec(SRC_spec_setup)
    SRC_spec_setup['wf_task_index'] = '_'.join(['setup', wf_task_index_prefix, str(current_task_index)])
    setup_task = task_class(task_input, is_autoparal=True, use_SRC_scheme=True, deps=deps, task_type=task_type)
    setup_fw = Firework(setup_task, spec=SRC_spec_setup, name=SRC_spec_setup['wf_task_index'])
    # Actual run of simulation
    SRC_spec_run = copy.deepcopy(SRC_spec)
    SRC_spec_run['wf_task_index'] = '_'.join(['run', wf_task_index_prefix, str(current_task_index)])
    run_task = task_class(task_input, is_autoparal=False, use_SRC_scheme=True, deps=deps, task_type=task_type)
    run_fw = Firework(run_task, spec=SRC_spec_run, name=SRC_spec_run['wf_task_index'])
    # Check memory firework
    SRC_spec_check = copy.deepcopy(SRC_spec)
    SRC_spec_check = set_short_single_core_to_spec(SRC_spec_check)
    SRC_spec_check['wf_task_index'] = '_'.join(['check', wf_task_index_prefix, str(current_task_index)])
    check_task = CheckTask(handlers=handlers, validators=validators)
    SRC_spec_check['_allow_fizzled_parents'] = True
    check_fw = Firework(check_task, spec=SRC_spec_check, name=SRC_spec_check['wf_task_index'])
    links_dict = {setup_fw.fw_id: [run_fw.fw_id],
                  run_fw.fw_id: [check_fw.fw_id]}
    return {'setup_fw': setup_fw, 'run_fw': run_fw, 'check_fw': check_fw, 'links_dict': links_dict,
            'fws': [setup_fw, run_fw, check_fw]}


@explicit_serialize
class FinalCleanUpTask(FireTaskBase):
    """
    Eliminates all the files in outdata based on the extension of the files and deletes of the files in
    indata and tmpdata.
    """
    task_type = 'finalclnup'

    def __init__(self, out_exts=None):
        """
        Args:
            out_exts: list of extensions. can be a list or a single string with extensions separated by a ",".
                If None "WFK" and "1WF" will be deleted
        """
        if out_exts is None:
            out_exts = ["WFK", "1WF"]
        if isinstance(out_exts, str):
            out_exts = [s.strip() for s in out_exts.split(',')]

        self.out_exts = out_exts

    @serialize_fw
    def to_dict(self):
        return dict(out_exts=self.out_exts)

    @classmethod
    def from_dict(cls, m_dict):
        return cls(out_exts=m_dict['out_exts'])

    @staticmethod
    def delete_files(d, exts=None):
        if exts and not isinstance(exts, (list, tuple)):
            exts = [exts]
        deleted_files = []
        if os.path.isdir(d):
            for f in os.listdir(d):
                if exts is None or "*" in exts or any(ext in f for ext in exts):
                    fp = os.path.join(d, f)
                    try:
                        if os.path.isfile(fp):
                            os.unlink(fp)
                        elif os.path.isdir(fp):
                            shutil.rmtree(fp)
                        elif os.path.islink(fp):
                            os.unlink(fp)
                        deleted_files.append(fp)
                    except:
                        logger.warning("Couldn't delete {}: {}".format(fp, traceback.format_exc()))

        return deleted_files

    def run_task(self, fw_spec):
        lp, fw_id = get_lp_and_fw_id_from_task(self, fw_spec)

        wf = lp.get_wf_by_fw_id_lzyfw(fw_id)

        deleted_files = []
        # iterate over all the fws and launches
        for fw_id, fw in wf.id_fw.items():
            for l in fw.launches+fw.archived_launches:
                l_dir = l.launch_dir

                deleted_files.extend(self.delete_files(os.path.join(l_dir, TMPDIR_NAME)))
                deleted_files.extend(self.delete_files(os.path.join(l_dir, INDIR_NAME)))
                deleted_files.extend(self.delete_files(os.path.join(l_dir, OUTDIR_NAME), self.out_exts))

        logging.info("Deleted files:\n {}".format("\n".join(deleted_files)))

        return FWAction(stored_data={'deleted_files': deleted_files})


@explicit_serialize
class DatabaseInsertTask(FireTaskBase):
    task_type = 'dbinsert'

    def __init__(self, insertion_data=None, criteria=None):
        if insertion_data is None:
            insertion_data = {'structure': 'get_final_structure_and_history'}
        self.insertion_data = insertion_data
        self.criteria = criteria

    @serialize_fw
    def to_dict(self):
        return dict(insertion_data=self.insertion_data, criteria=self.criteria)

    @classmethod
    def from_dict(cls, m_dict):
        return cls(insertion_data=m_dict['insertion_data'],
                   criteria=m_dict['criteria'] if 'criteria' in m_dict else None)

    @staticmethod
    def insert_objects():
        return None

    def run_task(self, fw_spec):
        lp, fw_id = get_lp_and_fw_id_from_task(self, fw_spec)

        wf = lp.get_wf_by_fw_id(fw_id)
        wf_module = importlib.import_module(wf.metadata['workflow_module'])
        wf_class = getattr(wf_module, wf.metadata['workflow_class'])

        database = fw_spec['mongo_database']
        if self.criteria is not None:
            entry = database.get_entry(criteria=self.criteria)
        else:
            entry = {}

        inserted = []
        for root_key, method_name in self.insertion_data.items():
            get_results_method = getattr(wf_class, method_name)
            results = get_results_method(wf)
            for key, val in results.items():
                entry[key] = jsanitize(val)
                inserted.append(key)

        if self.criteria is not None:
            database.save_entry(entry=entry)
        else:
            database.insert_entry(entry=entry)

        logging.info("Inserted data:\n{}".format('- {}\n'.join(inserted)))
        return FWAction()


@explicit_serialize
class MongoEngineDBInsertionTask(FireTaskBase):
    """
    Task that insert the results of a workflow in a database. The method calls the "get_mongoengine_results"
    of the workflow generator.
    The database is defined according to a DatabaseData.
    """

    def __init__(self, db_data, object_id=None):
        self.db_data = db_data
        #TODO allow for updates based on object_id. A bit tricky as one need to know the class of the Document
        # before calling get_results. (add object_id to dict methods)
        self.object_id = object_id

    @serialize_fw
    def to_dict(self):
        return dict(db_data=self.db_data.as_dict())

    @classmethod
    def from_dict(cls, m_dict):
        return cls(db_data=DatabaseData.from_dict(m_dict['db_data']))

    def run_task(self, fw_spec):
        self.db_data.connect_mongoengine()

        lp, fw_id = get_lp_and_fw_id_from_task(self, fw_spec)

        wf = lp.get_wf_by_fw_id_lzyfw(fw_id)
        wf_module = importlib.import_module(wf.metadata['workflow_module'])
        wf_class = getattr(wf_module, wf.metadata['workflow_class'])

        get_results_method = getattr(wf_class, 'get_mongoengine_results')

        #TODO extend for multiple documents?
        document = get_results_method(wf)

        with self.db_data.switch_collection(document.__class__) as document.__class__:
            #TODO it would be better to try to remove automatically the FileFields already saved if the save of
            # the document fails.
            document.save()


@explicit_serialize
class CheckMemoryTask(FireTaskBase):
    task_type = 'checkmem'

    def __init__(self, memory_increase_megabytes=1000, max_memory_megabytes=7600):
        self.memory_increase_megabytes = memory_increase_megabytes
        self.max_memory_megabytes = max_memory_megabytes

    @serialize_fw
    def to_dict(self):
        return dict(memory_increase_megabytes=self.memory_increase_megabytes,
                    max_memory_megabytes=self.max_memory_megabytes)

    @classmethod
    def from_dict(cls, m_dict):
        return cls(memory_increase_megabytes=m_dict['memory_increase_megabytes'],
                   max_memory_megabytes=m_dict['max_memory_megabytes'])

    # def run_task(self, fw_spec):
    #
    #     # Get the fw_id and launchpad
    #     if '_add_launchpad_and_fw_id' in fw_spec:
    #         lp = self.launchpad
    #         fw_id = self.fw_id
    #     else:
    #         try:
    #             fw_dict = loadfn('FW.json')
    #         except IOError:
    #             try:
    #                 fw_dict = loadfn('FW.yaml')
    #             except IOError:
    #                 raise RuntimeError("Launchpad/fw_id not present in spec and No FW.json nor FW.yaml file present: "
    #                                    "impossible to determine fw_id")
    #         lp = LaunchPad.auto_load()
    #         fw_id = fw_dict['fw_id']
    #
    #     # Treat the case where there was no memory error => forward "needed" outputs of the previous firework to the
    #     # next one.
    #     if not '_fizzled_parents' in fw_spec:
    #         stored_data = {}
    #         update_spec = {}
    #         mod_spec = []
    #         for task_type, task_info in fw_spec['previous_fws'].items():
    #             mod_spec.append({'_push_all': {'previous_fws->'+task_type: task_info}})
    #         return FWAction(stored_data=stored_data, update_spec=update_spec, mod_spec=mod_spec)
    #
    #     if len(fw_spec['_fizzled_parents']) > 1:
    #         raise ValueError('Multiple parents fizzled ... Only one is allowed.')
    #
    #     fizzled_fw_id = fw_spec['_fizzled_parents'][0]['fw_id']
    #     fizzled_fw = lp.get_fw_by_id(fizzled_fw_id)
    #     fizzled_fw_dir = fizzled_fw.launches[-1].launch_dir
    #
    #     # Analyze the stderr and stdout files of the resource manager system.
    #     qerr_info = None
    #     qout_info = None
    #     qerr_file = os.path.join(fizzled_fw_dir, 'queue.qerr')
    #     qout_file = os.path.join(fizzled_fw_dir, 'queue.qout')
    #     runerr_file = fw_spec['runerr_file'] if 'runerr_file' in fw_spec else None
    #     if os.path.exists(qerr_file):
    #         with open(qerr_file, "r") as f:
    #             qerr_info = f.read()
    #     if os.path.exists(qout_file):
    #         with open(qout_file, "r") as f:
    #             qout_info = f.read()
    #
    #     if qerr_info or qout_info:
    #         from pymatgen.io.abinit.scheduler_error_parsers import get_parser
    #         qtk_qadapter = fizzled_fw.spec['qtk_queueadapter']
    #         qtype = qtk_qadapter.QTYPE
    #         scheduler_parser = get_parser(qtype, err_file=qerr_file,
    #                                       out_file=qout_file, run_err_file=runerr_file)
    #
    #         if scheduler_parser is None:
    #             raise ValueError('Cannot find scheduler_parser for qtype {}'.format(qtype))
    #
    #         scheduler_parser.parse()
    #         queue_errors = scheduler_parser.errors
    #
    #         if queue_errors:
    #             # the queue errors in the task
    #             logger.debug('scheduler errors found:')
    #             logger.debug(str(queue_errors))
    #         else:
    #             if len(qerr_info) > 0:
    #                 logger.debug('found unknown queue error: {}'.format(str(qerr_info)))
    #                 raise ValueError(qerr_info)
    #                 # The job is killed or crashed but we don't know what happened
    #
    #         to_be_corrected = False
    #         for error in queue_errors:
    #             if isinstance(error, MemoryCancelError):
    #                 logger.debug('found memory error.')
    #                 to_be_corrected = True
    #         if to_be_corrected:
    #             if len(fizzled_fw.tasks) > 1:
    #                 raise ValueError('More than 1 task found in "memory-fizzled" firework, not yet supported')
    #             logger.debug('adding SRC detour')
    #             mytask = fizzled_fw.tasks[0]
    #             task_class = mytask.__class__
    #             # TODO: make this more general ... right now, it is based on AbinitInput and thus is strongly tight
    #             #       to abinit
    #             task_input = AbinitInput.from_dict(fizzled_fw.spec['_tasks'][0]['abiinput'])
    #             spec = fizzled_fw.spec
    #             initialization_info = fizzled_fw.spec['initialization_info']
    #             # Update the task index
    #             fizzled_fw_task_index = int(fizzled_fw.spec['wf_task_index'].split('_')[-1])
    #             new_index = fizzled_fw_task_index + 1
    #             # Update the memory in the queue adapter
    #             old_mem = qtk_qadapter.mem_per_proc
    #             new_mem = old_mem + self.memory_increase_megabytes
    #             if new_mem > self.max_memory_megabytes:
    #                 raise ValueError('New memory {:d} is larger than '
    #                                  'max memory per proc {:d}'.format(new_mem, self.max_memory_megabytes))
    #             qtk_qadapter.set_mem_per_proc(new_mem)
    #             spec['qtk_queueadapter'] = qtk_qadapter
    #             qadapter_spec = qtk_qadapter.get_subs_dict()
    #             spec['_queueadapter'] = qadapter_spec
    #             if 'run_timelimit' in spec:
    #                 run_timelimit = spec['run_timelimit']
    #             else:
    #                 run_timelimit = None
    #
    #             SRC_fws = SRCFireworks(task_class=task_class, task_input=task_input, spec=spec,
    #                                    initialization_info=initialization_info,
    #                                    wf_task_index_prefix=spec['wf_task_index_prefix'],
    #                                    current_task_index=new_index,
    #                                    current_memory_per_proc_mb=new_mem,
    #                                    memory_increase_megabytes=self.memory_increase_megabytes,
    #                                    max_memory_megabytes=self.max_memory_megabytes,
    #                                    task_type=mytask.task_type, run_timelimit=run_timelimit)
    #             wf = Workflow(fireworks=SRC_fws['fws'], links_dict=SRC_fws['links_dict'])
    #             return FWAction(detours=[wf])
    #     raise ValueError('Could not check for memory problem ...')


@explicit_serialize
class CheckTask(FireTaskBase):
    task_type = 'check'

    optional_params = ['handlers', 'validators']

    def __init__(self, handlers=None, validators=None, max_restarts=10):
        #super(CheckTask).__init__(self)
        self.handlers = handlers
        # Check that there is only one FIRST and one LAST handler (PRIORITY_FIRST and PRIORITY_LAST)
        if self.handlers is not None:
            h_priorities = [h.handler_priority for h in self.handlers]
            nhfirst = h_priorities.count(SRCErrorHandler.PRIORITY_FIRST)
            nhlast = h_priorities.count(SRCErrorHandler.PRIORITY_LAST)
            if nhfirst > 1 or nhlast > 1:
                raise ValueError('More than one first or last handler :\n'
                                 ' - nfirst : {:d}\n - nlast : {:d}'.format(nhfirst,
                                                                            nhlast))
        self.validators = validators
        self.max_restarts = max_restarts

    @serialize_fw
    def to_dict(self):
        return dict(handlers=[h.as_dict() for h in self.handlers] if self.handlers is not None else None,
                    validators=[v.as_dict() for v in self.validators] if self.validators is not None else None,
                    max_restarts=self.max_restarts)

    @classmethod
    def from_dict(cls, m_dict):
        m = MontyDecoder()
        return cls(handlers=m.process_decoded(m_dict['handlers']), validators=m.process_decoded(m_dict['validators'])
                   , max_restarts=m_dict['max_restarts'])

    def run_task(self, fw_spec):
        # Get the fw_id and launchpad
        if '_add_launchpad_and_fw_id' in fw_spec:
            lp = self.launchpad
            fw_id = self.fw_id
        else:
            try:
                with open('FW.json', "rt") as fh:
                    fw_dict = json.load(fh, cls=MontyDecoder)
            except IOError:
                try:
                    with open('FW.yaml', "rt") as fh:
                        fw_dict = yaml.load(fh)
                except IOError:
                    raise RuntimeError("Launchpad/fw_id not present in spec and No FW.json nor FW.yaml file present: "
                                       "impossible to determine fw_id")
            lp = LaunchPad.auto_load()
            fw_id = fw_dict['fw_id']

        # Treat the case where there was some error that led to a fizzled state
        if '_fizzled_parents' in fw_spec:
            if len(fw_spec['_fizzled_parents']) != 1:
                raise ValueError('CheckTask\'s Firework should have exactly one parent firework')
            # Get the fizzled fw
            fizzled_fw_id = fw_spec['_fizzled_parents'][0]['fw_id']
            fizzled_fw = lp.get_fw_by_id(fizzled_fw_id)
            # Sort handlers by their priority
            sorted_handlers = sorted([h for h in self.handlers if h.allow_fizzled], key=lambda x: x.handler_priority)
            # Get the corrections for all the handlers
            corrections = []
            for handler in sorted_handlers:
                # Set needed data for the handlers (the spec of this check task/fw and the fw that has to be checked)
                handler.src_setup(fw_spec=fw_spec, fw_to_check=fizzled_fw)
                if handler.check():
                    corrections.append(handler.correct())
                    if handler.skip_remaining_handlers:
                        break

            # In case of a fizzled parent, at least one correction is needed !
            if len(corrections) == 0:
                raise RuntimeError('No corrections found for fizzled firework ...')

            # Apply the corrections
            fw_action = self.apply_corrections(fw_to_correct=fizzled_fw, corrections=corrections)
            return fw_action
        # Treat the case where there was no fizzled parents => forward "needed" outputs of the previous firework to the
        # next one.
        else:
            # Get the previous fw
            this_lzy_wf = lp.get_wf_by_fw_id_lzyfw(fw_id)
            parents_fw_ids = this_lzy_wf.links.parent_links[fw_id]
            if len(parents_fw_ids) != 1:
                raise ValueError('CheckTask\'s Firework should have exactly one parent firework')
            run_fw = lp.get_fw_by_id(parents_fw_ids[0])
            # Get the corrections for all the handlers
            # Sort handlers by their priority
            if self.handlers is not None:
                sorted_handlers = sorted([h for h in self.handlers if h.allow_completed],
                                         key=lambda x: x.handler_priority)
            else:
                sorted_handlers = []
            # Get the corrections for all the handlers
            corrections = []
            for handler in sorted_handlers:
                # Set needed data for the handlers (the spec of this check task/fw and the fw that has to be checked)
                handler.src_setup(fw_spec=fw_spec, fw_to_check=run_fw)
                if handler.check():
                    corrections.append(handler.correct())
                if handler.skip_remaining_handlers:
                    break

            # If some corrections are found, apply and return the FWAction
            if len(corrections) > 0:
                fw_action = self.apply_corrections(fw_to_correct=run_fw, corrections=corrections)
                return fw_action

            # Validate the results if no error was found
            validators = self.validators if self.validators is not None else []
            for validator in validators:
                if not validator.check():
                    raise RuntimeError('Validator invalidate results ...')
            stored_data = {}
            update_spec = {}
            mod_spec = []
            for task_type, task_info in fw_spec['previous_fws'].items():
                mod_spec.append({'_push_all': {'previous_fws->'+task_type: task_info}})
            return FWAction(stored_data=stored_data, update_spec=update_spec, mod_spec=mod_spec)

    def apply_corrections(self, fw_to_correct, corrections):
        # Apply the corrections
        spec = fw_to_correct.spec
        modder = Modder()
        for correction in corrections:
            actions = correction['actions']
            for action in actions:
                if action['action_type'] == 'modify_object':
                    if action['object']['source'] == 'fw_spec':
                        myobject = spec[action['object']['key']]
                    else:
                        raise NotImplementedError('Object source "{}" not implemented in '
                                                  'CheckTask'.format(action['object']['source']))
                    newobj = modder.modify_object(action['action'], myobject)
                    spec[action['object']['key']] = newobj
                elif action['action_type'] == 'modify_dict':
                    if action['dict']['source'] == 'fw_spec':
                        mydict = spec[action['dict']['key']]
                    else:
                        raise NotImplementedError('Dict source "{}" not implemented in '
                                                  'CheckTask'.format(action['dict']['source']))
                    modder.modify(action['action'], mydict)
                else:
                    raise NotImplementedError('Action type "{}" not implemented in '
                                              'CheckTask'.format(action['action_type']))
        # Keep track of the corrections that have been applied
        spec['SRC_check_corrections'] = corrections

        # Update the task index
        fws_task_index = int(fw_to_correct.spec['wf_task_index'].split('_')[-1])
        new_index = fws_task_index + 1
        # Update the Fireworks _queueadapter key
        #TODO: in the future, see whether the FW queueadapter might be replaced by the qtk_queueadapter ?
        #      ... to be discussed with Anubhav, when the qtk queueadapter is in a qtk toolkit and not anymore
        #          in pymatgen/io/abinit
        spec['_queueadapter'] = spec['qtk_queueadapter'].get_subs_dict()
        queue_adapter_update = get_queue_adapter_update(qtk_queueadapter=spec['qtk_queueadapter'],
                                                        corrections=corrections)

        # Get and update the task_input if needed
        # TODO: make this more general ... right now, it is based on AbinitInput and thus is strongly tight
        #       to abinit due to abiinput, deps, ...
        mytask = fw_to_correct.tasks[0]
        task_class = mytask.__class__
        decoder = MontyDecoder()
        task_input = decoder.process_decoded(fw_to_correct.spec['_tasks'][0]['abiinput'])
        initialization_info = fw_to_correct.spec['initialization_info']
        deps = mytask.deps

        # Create the new Setup/Run/Check fireworks
        SRC_fws = createSRCFireworksOld(task_class=task_class, task_input=task_input, SRC_spec=spec,
                                        initialization_info=initialization_info,
                                        wf_task_index_prefix=spec['wf_task_index_prefix'],
                                        current_task_index=new_index,
                                        handlers=self.handlers, validators=self.validators,
                                        deps=deps,
                                        task_type=mytask.task_type, queue_adapter_update=queue_adapter_update)
        wf = Workflow(fireworks=SRC_fws['fws'], links_dict=SRC_fws['links_dict'])
        return FWAction(detours=[wf])


def print_myself():
    print('myself')

def get_fw_task_manager(fw_spec):
    if 'ftm_file' in fw_spec:
        ftm = FWTaskManager.from_file(fw_spec['ftm_file'])
    else:
        ftm = FWTaskManager.from_user_config()
    ftm.update_fw_policy(fw_spec.get('fw_policy', {}))
    return ftm


def apply_corrections_to_spec(corrections, spec):
    modder = Modder()
    for correction in corrections:
        actions = correction['actions']
        for action in actions:
            if action['action_type'] == 'modify_object':
                if action['object']['source'] == 'fw_spec':
                    myobject = spec[action['object']['key']]
                else:
                    raise NotImplementedError('Object source "{}" not implemented in '
                                              'CheckTask'.format(action['object']['source']))
                newobj = modder.modify_object(action['action'], myobject)
                spec[action['object']['key']] = newobj
            elif action['action_type'] == 'modify_dict':
                if action['dict']['source'] == 'fw_spec':
                    mydict = spec[action['dict']['key']]
                else:
                    raise NotImplementedError('Dict source "{}" not implemented in '
                                              'CheckTask'.format(action['dict']['source']))
                modder.modify(action['action'], mydict)
            else:
                raise NotImplementedError('Action type "{}" not implemented in '
                                          'CheckTask'.format(action['action_type']))


def get_queue_adapter_update(qtk_queueadapter, corrections, qa_params=None):
    if qa_params is None:
        qa_params = ['timelimit', 'mem_per_proc', 'master_mem_overhead']
    queue_adapter_update = {}
    for qa_param in qa_params:
        if qa_param == 'timelimit':
            queue_adapter_update[qa_param] = qtk_queueadapter.timelimit
        elif qa_param == 'mem_per_proc':
            queue_adapter_update[qa_param] = qtk_queueadapter.mem_per_proc
        elif qa_param == 'master_mem_overhead':
            queue_adapter_update[qa_param] = qtk_queueadapter.master_mem_overhead
        else:
            raise ValueError('Wrong queue adapter parameter for update')
    for correction in corrections:
        for action in correction['actions']:
            if action['object']['key'] == 'qtk_queueadapter':
                qa_update = action['action']['_set']
                queue_adapter_update.update(qa_update)
    return queue_adapter_update



# coding: utf-8
"""
Utilities to get factors of a given number or related things ...
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import itertools
import math

def lowest_nn_gte_mm(mm, factors):
    if mm < 1:
        raise ValueError('Number mm should be >= 1')
    if not any([factor > 1 for factor in factors]):
        raise ValueError('At least one factor should be > 1')
    if len(set(factors)) != len(factors):
        raise ValueError('The factors should all be different')
    lowest = -1
    current_exponents = -1
    max_exponents_factors = [int(math.ceil(math.log(float(mm), factor))) for factor in factors]

    exponents_possibilities = [range(max_exponent+1) for max_exponent in max_exponents_factors]
    for exponents in itertools.product(*exponents_possibilities):
        nn = 1
        for ifactor, factor in enumerate(factors):
            nn *= factor ** exponents[ifactor]
        if nn < mm:
            continue
        elif nn == mm:
            lowest = nn
            current_exponents = exponents
            break
        else:
            if lowest == -1:
                lowest = nn
                current_exponents = exponents
            elif nn < lowest:
                lowest = nn
                current_exponents = exponents

    return lowest, current_exponents


from __future__ import print_function, division, unicode_literals, absolute_import

import pytest
import os
import glob
import tempfile
import filecmp
import numpy.testing.utils as nptu

from abipy.abio.input_tags import *
from abipy.abio.factories import dfpt_from_gsinput
from abipy.core.testing import has_abinit
from fireworks.core.rocket_launcher import rapidfire
from abiflows.fireworks.workflows.abinit_workflows import DfptFWWorkflow
from abiflows.fireworks.utils.fw_utils import get_fw_by_task_index, load_abitask
from abiflows.core.testing import AbiflowsIntegrationTest, check_restart_task_type

ABINIT_VERSION = "8.6.1"

# pytestmark = [pytest.mark.skipif(not has_abinit(ABINIT_VERSION), reason="Abinit version {} is not in PATH".format(ABINIT_VERSION)),
#               pytest.mark.skipif(not has_fireworks(), reason="fireworks paackage is missing"),
#               pytest.mark.skipif(not has_mongodb(), reason="no connection to mongodb")]

pytestmark = pytest.mark.usefixtures("cleandb")


class ItestDfpt(AbiflowsIntegrationTest):

    def itest_dfpt_full(self, lp, fworker, tmpdir, input_scf_phonon_gan_low, use_autoparal, db_data):
        """
        Simple test of DteFWWorkflow with autoparal True and False.
        Doesn't run anaddb since anaddb does not support a DDB with third order perturbations along with
        other perturbations.
        Skips dte permutations.
        """

        # dte calculations only work with selected values of ixc
        input_scf_phonon_gan_low['ixc'] = 7
        dfpt_inputs = dfpt_from_gsinput(input_scf_phonon_gan_low, ph_ngqpt=[2, 2, 2], do_ddk=True, do_dde=True,
                                        do_strain=True, do_dte=True, skip_dte_permutations=True,
                                        ddk_tol = {"tolwfr": 1.0e-16}, dde_tol = {"tolvrs": 1.0e-7},
                                        strain_tol={"tolvrs": 1.0e-7}, ph_tol={"tolvrs": 1.0e-7})

        wf = DfptFWWorkflow(input_scf_phonon_gan_low, ddk_inp = dfpt_inputs.filter_by_tags(DDK),
                           dde_inp = dfpt_inputs.filter_by_tags(DDE), strain_inp=dfpt_inputs.filter_by_tags(STRAIN),
                           ph_inp = dfpt_inputs.filter_by_tags(PH_Q_PERT), dte_inp = dfpt_inputs.filter_by_tags(DTE),
                            nscf_inp=dfpt_inputs.filter_by_tags(NSCF),initialization_info={"kppa": 100},
                            autoparal=use_autoparal)

        wf.add_mongoengine_db_insertion(db_data)
        wf.add_final_cleanup(["WFK"])

        scf_fw_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        # check the effect of the final cleanup
        scf_task = load_abitask(get_fw_by_task_index(wf, "scf", index=1))

        assert len(glob.glob(os.path.join(scf_task.outdir.path, "*_WFK"))) == 0
        assert len(glob.glob(os.path.join(scf_task.outdir.path, "*_DEN"))) == 1
        assert len(glob.glob(os.path.join(scf_task.tmpdir.path, "*"))) == 0
        assert len(glob.glob(os.path.join(scf_task.indir.path, "*"))) == 0

    def itest_dfpt_anaddb_ph(self, lp, fworker, tmpdir, input_scf_phonon_gan_low, db_data):
        """
        Simple test of DteFWWorkflow with autoparal True and False.
        Skips dte permutations.
        """

        dfpt_inputs = dfpt_from_gsinput(input_scf_phonon_gan_low, ph_ngqpt=[2, 2, 2], do_ddk=True, do_dde=True,
                                        do_strain=True, do_dte=False,
                                        ddk_tol = {"tolwfr": 1.0e-16}, dde_tol = {"tolvrs": 1.0e-7},
                                        strain_tol={"tolvrs": 1.0e-7}, ph_tol={"tolvrs": 1.0e-7})

        wf = DfptFWWorkflow(input_scf_phonon_gan_low, ddk_inp = dfpt_inputs.filter_by_tags(DDK),
                           dde_inp = dfpt_inputs.filter_by_tags(DDE), strain_inp=dfpt_inputs.filter_by_tags(STRAIN),
                           ph_inp = dfpt_inputs.filter_by_tags(PH_Q_PERT), dte_inp = dfpt_inputs.filter_by_tags(DTE),
                            nscf_inp=dfpt_inputs.filter_by_tags(NSCF),initialization_info={"kppa": 100},
                            autoparal=False)

        wf.add_anaddb_dfpt_fw(input_scf_phonon_gan_low.structure, ph_ngqpt=[2, 2, 2], nqsmall=2, ndivsm=3)
        wf.add_mongoengine_db_insertion(db_data)
        wf.add_final_cleanup(["WFK"])

        scf_fw_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        scf_task = load_abitask(get_fw_by_task_index(wf, "scf", index=1))

        # check the save in the DB
        from abiflows.database.mongoengine.abinit_results import DfptResult
        with db_data.switch_collection(DfptResult) as DteResult:
            results = DteResult.objects()
            assert len(results) == 1
            r = results[0]

            assert r.abinit_input.structure.to_mgobj() == input_scf_phonon_gan_low.structure
            assert r.abinit_output.structure.to_mgobj() == input_scf_phonon_gan_low.structure
            assert r.abinit_input.ecut == input_scf_phonon_gan_low['ecut']
            assert r.abinit_input.kppa == 100
            nptu.assert_array_equal(r.abinit_input.gs_input.to_mgobj()['ngkpt'], input_scf_phonon_gan_low['ngkpt'])

            ana_task = load_abitask(get_fw_by_task_index(wf, "anaddb", index=None))

            with tempfile.NamedTemporaryFile(mode="wb") as db_file:
                db_file.write(r.abinit_output.anaddb_nc.read())
                db_file.seek(0)
                assert filecmp.cmp(ana_task.anaddb_nc_path, db_file.name)

            mrgddb_task = load_abitask(get_fw_by_task_index(wf, "mrgddb", index=None))

            with tempfile.NamedTemporaryFile(mode="wb") as db_file:
                db_file.write(r.abinit_output.ddb.read())
                db_file.seek(0)
                assert filecmp.cmp(mrgddb_task.merged_ddb_path, db_file.name)

        if self.check_numerical_values:
            with scf_task.open_gsr() as gsr:
                assert gsr.energy == pytest.approx(-680.402255069, rel=0.005)

            ana_task = load_abitask(get_fw_by_task_index(wf, "anaddb", index=None))
            with ana_task.open_anaddbnc() as ananc:
                assert float(ananc.eps0[0,0]) == pytest.approx(64.8276774889143, rel=0.15)

                e = ananc.elastic_data
                if has_abinit("8.9.3"):
                    assert float(e.elastic_relaxed[0,0,0,0]) == pytest.approx(41.230540749230556, rel=0.15)

    def itest_dfpt_anaddb_dte(self, lp, fworker, tmpdir, input_scf_phonon_gan_low, db_data):
        """
        Simple test of DteFWWorkflow with dte.
        Does not run anaddb due to problems when dealing with DDB file containing 3rd order derivatives and
        other terms.
        """

        # dte calculations only work with selected values of ixc
        input_scf_phonon_gan_low['ixc'] = 7
        dfpt_inputs = dfpt_from_gsinput(input_scf_phonon_gan_low, ph_ngqpt=[1,1,1], do_ddk=True, do_dde=True,
                                        do_strain=False, do_dte=True, skip_dte_permutations=True,
                                        ddk_tol = {"tolwfr": 1.0e-16}, dde_tol = {"tolvrs": 1.0e-7},
                                        strain_tol={"tolvrs": 1.0e-7}, ph_tol={"tolvrs": 1.0e-7})

        wf = DfptFWWorkflow(input_scf_phonon_gan_low, ddk_inp = dfpt_inputs.filter_by_tags(DDK),
                            dde_inp = dfpt_inputs.filter_by_tags(DDE), strain_inp=dfpt_inputs.filter_by_tags(STRAIN),
                            ph_inp = dfpt_inputs.filter_by_tags(PH_Q_PERT), dte_inp = dfpt_inputs.filter_by_tags(DTE),
                            nscf_inp=dfpt_inputs.filter_by_tags(NSCF),initialization_info={"kppa": 100},
                            autoparal=False)

        wf.add_anaddb_dfpt_fw(input_scf_phonon_gan_low.structure)
        wf.add_mongoengine_db_insertion(db_data)
        wf.add_final_cleanup(additional_spec={"test_spec":1})

        scf_fw_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        if self.check_numerical_values:
            ana_task = load_abitask(get_fw_by_task_index(wf, "anaddb", index=None))
            with ana_task.open_anaddbnc() as ananc:
                assert float(ananc.dchide[0,0,2]) == pytest.approx(-1.69328765210, rel=0.15)

    def itest_not_converged(self, lp, fworker, tmpdir, input_scf_phonon_si_low):
        """
        Tests the missed convergence and restart for the Strain task. Other restarts are checked in
        the PhononWorkflow tests.
        """

        # only strain perturbations included
        dfpt_inputs = dfpt_from_gsinput(input_scf_phonon_si_low, do_ddk=False, do_dde=False,
                                        do_strain=True, do_dte=False, skip_dte_permutations=True,
                                        strain_tol={"tolvrs": 1.0e-7} )
        dfpt_inputs.set_vars(nstep=3)
        wf = DfptFWWorkflow(input_scf_phonon_si_low, ddk_inp=None, dde_inp=None, ph_inp=None,
                            strain_inp=dfpt_inputs.filter_by_tags(STRAIN), dte_inp=None,
                            nscf_inp=None, initialization_info=None, autoparal=False)

        scf_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_id = old_new[scf_id]

        # run the scf
        rapidfire(lp, fworker, m_dir=str(tmpdir), nlaunches=1)
        # pause all the remaining workflow and reignite the task types one by one to check the restart
        lp.pause_wf(scf_id)

        # strain
        check_restart_task_type(lp, fworker, tmpdir, scf_id, "strain_pert_0")
